from .fcn import (
    FCN8s,
    FCN16s,
    FCN32s,
    VggFCN8s,
    VggFCN16s,
    VggFCN32s,
    VggFCNs,
    VggFCN,
    ResFCN,
    DenseFCN,
    MobileFCN,
    MobileFCN1
)

from .unet import UNet



"""
Copyright (c) 2023, Auorui.
All rights reserved.
"""
import cv2
from pylab import *
from skimage.morphology import disk
from skimage.filters import rank
from skimage import measure
from pyzjr.augmentation.mask_ops import BinaryImg
import pyzjr.Z as Z

def getContours(img, cThr=(100, 100), minArea=1000, filter=0, draw=True):
    """
    :param img: 输入图像
    :param cThr: 阈值
    :param minArea: 更改大小
    :param filter: 过滤
    :param draw: 绘制边缘
    :return: 返回图像轮廓
    """
    imgGray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    imgBlur = cv2.GaussianBlur(imgGray, (5, 5), 1)
    imgCanny = cv2.Canny(imgBlur, cThr[0], cThr[1])
    kernel = np.ones((5, 5))
    imgDial = cv2.dilate(imgCanny, kernel, iterations=3)
    imgThre = cv2.erode(imgDial, kernel, iterations=2)
    contours, _ = cv2.findContours(imgThre, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    finalCountours = []
    for i in contours:
        area = cv2.contourArea(i)
        if area > minArea:
            peri = cv2.arcLength(i, True)
            approx = cv2.approxPolyDP(i, 0.02 * peri, True)
            bbox = cv2.boundingRect(approx)
            if filter > 0:
                if len(approx) == filter:
                    finalCountours.append([len(approx), area, approx, bbox, i])
            else:
                finalCountours.append([len(approx), area, approx, bbox, i])
    finalCountours = sorted(finalCountours, key=lambda x: x[1], reverse=True)
    if draw:
        for con in finalCountours:
            cv2.drawContours(img, con[4], -1, (0, 0, 255), 3)
    return img, finalCountours

def labelpoint(im, click=4):
    """
    交互式标注
    :param im: 图像,采用im = Image.open(?)的方式打开，颜色空间才是正常的
    :param click: 点击次数、默认为4
    :return: 返回点的坐标
    """
    imshow(im)
    print(f'please click {click} points')
    x = ginput(click)
    print('you clicked:',x)
    return x

def get_warp_perspective(img, targetWH):
    """
    标注方式按照先上后下，先左后右的顺序
    :param img: 图像
    :param targetWH: 已知目标物体的宽度,长度
    :return: 修正后的图像
    """
    width, height = targetWH
    pts1 = np.float32(labelpoint(img))
    pts2 = np.float32([[0, 0], [width, 0], [0, height], [width, height]])
    matrix = cv2.getPerspectiveTransform(pts1, pts2)
    imgOutput = cv2.warpPerspective(img, matrix, (width, height))
    return imgOutput

def SearchOutline(binary_image):
    """
    在给定图像中搜索轮廓并返回轮廓的坐标列表。
    :param binary_image: 要搜索轮廓的图像。
    :return: 包含轮廓坐标的列表，每个坐标表示裂缝的一个点，坐标格式为 [(x, y),...]。
    """
    contours = measure.find_contours(binary_image, level=128, fully_connected='low', positive_orientation='low')
    contours_xy = [np.fliplr(np.vstack(contour)).astype(np.int32) for contour in contours]
    return contours_xy

def drawOutline(blackbackground, contours, color=Z.purple, thickness=1):
    """绘制轮廓"""
    cv2.drawContours(blackbackground, contours, -1, color, thickness=thickness)

def gradientOutline(img, radius=2):
    """
    对图像进行梯度边缘检测，并返回边缘强度图像。
    
    :param img: 输入的图像(内部有二值转化)
    :param radius: 半径,默认为2。
    :return: 返回经过梯度边缘检测处理后的边缘强度图像
    """
    image = BinaryImg(img)
    denoised = rank.median(image, disk(radius))
    gradient = rank.gradient(denoised, disk(radius))
    gradient = cv2.cvtColor(gradient, cv2.COLOR_GRAY2BGR)

    return gradient



"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is used for visualizing useful functions.
"""
import cv2
import time
import logging
import numpy as np
from functools import wraps

import pyzjr.Z as Z
from pyzjr.visualize.io import StackedImagesV1, VideoCap


class FPS:
    """
    Detect video frame rate and refresh the video display
    Examples:
    ```
        fpsReader = FPS()
        Vcap = VideoCap(mode=0)
        while True:
            img = Vcap.read()
            fps, img = fpsReader.update(img)
            Vcap.show("ss", img)
    ```
    """
    def __init__(self):
        self.pTime = time.time()

    def update(self, img=None, pos=(20, 50), color=(255, 0, 0), scale=3, thickness=3):
        """
        Update frame rate
        :param img: The displayed image can be left blank if only the fps value is needed
        :param pos: Position on FPS on image
        :param color: The color of the displayed FPS value
        :param scale: The proportion of displayed FPS values
        :param thickness: The thickness of the displayed FPS value
        """
        cTime = time.time()
        try:
            fps = 1 / (cTime - self.pTime)
            self.pTime = cTime
            if img is None:
                return fps
            else:
                cv2.putText(img, f'FPS: {int(fps)}', pos, cv2.FONT_HERSHEY_PLAIN,
                            scale, color, thickness)
                return fps, img
        except:
            return 0

class Timer:
    def __init__(self):
        """Start is called upon creation"""
        self.times = []
        self.start()
    def start(self):
        """initial time"""
        self.gap = time.time()
    def stop(self):
        """The time interval from the start of timing to calling the stop method"""
        self.times.append(time.time() - self.gap)
        return self.times[-1]
    def avg(self):
        """Average time consumption"""
        return sum(self.times) / len(self.times)
    def total(self):
        """Total time consumption"""
        return sum(self.times)
    def cumsum(self):
        """Accumulated sum of time from the previous n runs"""
        return np.array(self.times).cumsum().tolist()


class Runcodes:
    """
    Comparing the running time of different algorithms.
    Examples:
    ```
        with Runcodes("inference time"):
            output = ...
    ```
    """
    def __init__(self, description='Done'):
        self.description = description

    def __enter__(self):
        self.timer = Timer()
        return self

    def __exit__(self, *args):
        print(f'{self.description}: {self.timer.stop():.7f} sec')


def timing(decimal=5):
    """计时器装饰器，用于测量函数执行的时间。"""
    def decorator(function):
        @wraps(function)
        def timingwrap(*args, **kwargs):
            print(function.__name__, flush=True)
            start = time.perf_counter()
            res = function(*args, **kwargs)
            end = time.perf_counter()
            execution_time = end - start
            format_string = "{:.{}f}".format(execution_time, decimal)
            print(function.__name__, "delta time (s) =", format_string, flush=True)
            return res
        return timingwrap
    return decorator


class ColorFind():
    def __init__(self, trackBar=False, name="Bars"):
        self.trackBar = trackBar
        self.name = name
        if self.trackBar:
            self.initTrackbars()

    def empty(self, a):
        pass

    def initTrackbars(self):
        """
        :return:初始化轨迹栏
        """
        cv2.namedWindow(self.name)
        cv2.resizeWindow(self.name, 640, 240)
        cv2.createTrackbar("Hue Min", self.name, 0, 179, self.empty)
        cv2.createTrackbar("Hue Max", self.name, 179, 179, self.empty)
        cv2.createTrackbar("Sat Min", self.name, 0, 255, self.empty)
        cv2.createTrackbar("Sat Max", self.name, 255, 255, self.empty)
        cv2.createTrackbar("Val Min", self.name, 0, 255, self.empty)
        cv2.createTrackbar("Val Max", self.name, 255, 255, self.empty)

    def getTrackbarValues(self):
        """
        Gets the trackbar values in runtime
        :return: hsv values from the trackbar window
        """
        hmin = cv2.getTrackbarPos("Hue Min", self.name)
        smin = cv2.getTrackbarPos("Sat Min", self.name)
        vmin = cv2.getTrackbarPos("Val Min", self.name)
        hmax = cv2.getTrackbarPos("Hue Max", self.name)
        smax = cv2.getTrackbarPos("Sat Max", self.name)
        vmax = cv2.getTrackbarPos("Val Max", self.name)
        HsvVals = [[hmin, smin, vmin], [hmax, smax, vmax]]

        return HsvVals

    def protect_region(self, mask, threshold=None):
        """
        * 用于保护掩膜图的部分区域
        :param mask: 掩膜图
        :param threshold: 如果为None,则为不保护，如果是长为4的列表，则进行特定区域的保护
        :return: 返回进行保护区域的掩膜图

        example:    [0, img.shape[1], 0, img.shape[0]]为全保护状态，
                    x_start可以保护大于x的部分
                    x_end可以保护小于x的部分
                    y_start可以保护图像下方的部分
                    y_end可以保护图像上方的部分
        """
        if threshold == None:
            return mask
        else:
            x_start, x_end, y_start, y_end = threshold[:4]
            mask[y_start:y_end, x_start:x_end] = 0
            return mask

    def MaskZone(self, img, HsvVals):
        """
        * 生成掩膜图以及融合图像
        :param img: 输入图像
        :param HsvVals: 可以通过getTrackbarValues获得,也可调取Z.HSV的值
        :return: 返回融合图、掩膜图、HSV图
        """
        imgHSV = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
        lower = np.array(HsvVals[0])
        upper = np.array(HsvVals[1])
        mask = cv2.inRange(imgHSV, lower, upper)
        imgResult = cv2.bitwise_and(img, img, mask=mask)
        return imgResult, mask

    def update(self, img, myColor=None):
        """
        :param img: 需要在其中找到颜色的图像
        :param myColor: hsv上下限列表
        :return: mask带有检测到颜色的白色区域的roi图像
                 imgColor彩色图像仅显示检测到的区域
        """
        imgColor = [],
        mask = []
        if self.trackBar:
            myColor = self.getTrackbarValues()

        if isinstance(myColor, str):
            myColor = self.getColorHSV(myColor)

        if myColor is not None:
            imgColor, mask = self.MaskZone(img, myColor)
        return imgColor, mask

    def getColorHSV(self, myColor):
        if myColor == 'red':
            output = [[146, 141, 77], [179, 255, 255]]
        elif myColor == 'green':
            output = [[44, 79, 111], [79, 255, 255]]
        elif myColor == 'blue':
            output = [[103, 68, 130], [128, 255, 255]]
        else:
            output = None
            logging.warning("Color Not Defined")
            logging.warning("Available colors: red, green, blue ")

        return output

def DetectImageColor(img, ConsoleOut=True, threshold=None, scale=1.0):
    """
    * 轨迹栏检测图片,此函数仅仅作为使用示例
    :param img: 图片
    :param name: 轨迹栏名
    :param ConsoleOut: 用于是否控制台打印HsvVals的值
    :param threshold: 阈值，用于保护图片的区域
    :param scale: 规模大小
    :return:
    """
    ColF = ColorFind(True, "DetectImg")
    while True:
        imgHSV = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
        HsvVals = ColF.getTrackbarValues()
        if ConsoleOut:
            print(HsvVals)
        imgResult, mask = ColF.update(img,HsvVals)
        pro_mask = ColF.protect_region(mask, threshold)
        imgStack = StackedImagesV1(scale, ([img, imgHSV],[pro_mask,imgResult]))
        cv2.imshow("Stacked Images", imgStack)
        k = cv2.waitKey(1)
        if k == Z.Esc:
            break

def DetectVideoColor(mode=0, myColor=None, scale=1.0):
    """
    * 轨迹栏检测摄像头,此函数仅仅作为使用示例
    :param mode: 检测模式,默认本地摄像头,可传入video路径
    :param myColor: getColorHSV返回的一些测好的Hsv值
    :param scale: 规模大小
    """
    if myColor:
        Cf = False
    else:
        Cf = True
    Vcap = VideoCap(mode=mode)
    ColF = ColorFind(Cf, "DetectVideo")
    while True:
        img = Vcap.read()
        imgColor, mask = ColF.update(img, myColor)
        stackedimg = StackedImagesV1(scale, [img, imgColor])
        Vcap.show("DetectVideo", stackedimg)

if __name__=="__main__":
    # DetectVideo(myColor="red")
    @timing(decimal=5)
    def test_function():
        time.sleep(2.5)
    test_function()
    # imagePath = r"D:\PythonProject\pyzjr\pyzjr\test.png"
    # img = cv2.imread(imagePath)
    # DetectImageColor(img)


"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is applied to image reading and display.
"""
import os
import cv2
import sys
import imageio
import numpy as np
from PIL import Image
from pathlib import Path
from urllib import request
from matplotlib import pyplot as plt

plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'Times New Roman'
plt.rcParams['font.sans-serif'] = ['SimHei']
plt.rcParams['axes.unicode_minus'] = False
from pyzjr.utils.check import (is_file, is_gray_image, is_numpy, is_list, is_url, is_pil,
                               is_tensor)
from pyzjr.visualize.colorspace import bgr2rgb
from pyzjr.utils.converter import cv2pil

def imreader(filename, is_gray=False, flip=None, to_rgb=False):
    """
    :param filename: Path to the image file.
    :param is_gray: If True, read the image in grayscale.
    :param flip: 0: Flip vertically.
                1: Flip horizontally.
               -1: Flip both vertically and horizontally.
    :param to_rgb: Convert to RGB, default to False

    Examples:
    ```
        image_path = r'dog.png'
        gray_image = imreader(image_path, is_gray=True, to_rgb=True)
        display("ss", gray_image)
        imwriter("ss.png", gray_image)
    ```
    """
    if not is_file(filename):
        raise ValueError(f"The file {filename} does not exist.")
    if isinstance(filename, Path):
        filename = str(filename.resolve())
    flags = cv2.IMREAD_GRAYSCALE if is_gray else cv2.IMREAD_COLOR
    image = cv2.imdecode(np.fromfile(filename, np.uint8), flags)
    if flip is not None:
        assert flip in [-1, 0, 1], f"The 'flip' parameter must be -1, 0, or 1."
        image = cv2.flip(image, flip)
    if is_gray_image(image):
        image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB) if to_rgb else image
    else:
        image = bgr2rgb(image) if to_rgb else image
    return image

def imwriter(filename: str, img: np.ndarray, params=None):
    """Write the image to a file."""
    try:
        cv2.imencode(Path(filename).suffix, img, params)[1].tofile(filename)
        return True
    except Exception:
        return False

def display(winname, imgArray, scale=1.):
    """Displays an image in the specified window."""
    _imshow = cv2.imshow  # copy to avoid recursion errors
    if is_numpy(imgArray):
        height, width = imgArray.shape[:2]
        new_width = int(width * scale)
        new_height = int(height * scale)
        image = cv2.resize(imgArray, (new_width, new_height))
        _imshow(winname.encode('unicode_escape').decode(), image)
    elif is_list(imgArray):
        image = StackedImagesV1(scale, imgArray)
        _imshow(winname.encode('unicode_escape').decode(), image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()

def url2image(url):
    """Same usage as cv2.imread()"""
    if is_url(url):
        res = request.urlopen(url, timeout=3)
    else:
        raise ValueError("The current input parameter does not conform to the URL format")
    try:
        image = np.asarray(bytearray(res.read()), dtype="uint8")
        image = cv2.imdecode(image, cv2.IMREAD_UNCHANGED)
    except:
        print('Load read - Image timeout!')
        image = []
    h, w, c = image.shape
    if c == 4:
        image = image[:, :, :3]
    return image

def imattributes(image):
    """Retrieve image attributes"""
    if is_numpy(image):
        imshape = image.shape
        size = imshape[:2]
        dtype = image.dtype
        height, width = size

        if len(imshape) == 2:
            depth = image.itemsize * 8
        else:
            depth = image.itemsize * 8 * imshape[2]

        return {
            "shape": imshape,
            "size": size,
            "height": height,
            "width": width,
            "dtype": dtype,
            "depth": depth,
            "source": "NumPy"
        }
    elif is_pil(image):
        width, height = image.size
        mode = image.mode

        if mode == 'L':  # Grayscale
            depth = 8
        elif mode in ('RGB', 'RGBA'):
            depth = 8 * 3 if mode == 'RGB' else 8 * 4
        elif mode in ('I', 'F'):  # Integer or float (single channel)
            depth = image.getbands()[0].size * 8
        else:
            depth = "Unknown"  # Other modes are not handled here

        return {
            "shape": (height, width, len(image.getbands())),  # Simulate a NumPy shape
            "size": (width, height),
            "height": height,
            "width": width,
            "dtype": "PIL mode " + mode,  # PIL doesn't have a direct dtype equivalent
            "depth": depth,
            "source": "PIL"
        }
    else:
        return "Input is not a NumPy array or PIL Image."

def imshowplt(image, stacked_shape=None, title=None, cmap=None, save_path=None):
    """
    Display image and add title
    Examples:
        color_image = Image.open("dog.png")
        imshowplt([color_image,color_image], stacked_shape=(2,1), title='PIL Color Image', cmap=None)
    """
    import matplotlib
    matplotlib.use("TkAgg")
    if is_list(image):
        image = Stackedplt(image, stacked_shape)
    else:
        if is_pil(image):
            if image.mode == 'L' or (image.mode == 'LA' and image.getbands()[1] == 'L'):
                cmap = 'gray'
        elif is_gray_image(image):
            cmap = 'gray'
    plt.imshow(image, cmap=cmap)
    plt.axis('off')
    if title:
        plt.title(title)
    if save_path:
        plt.savefig(save_path)
    plt.show()

class VideoCap():
    """
    Customized Python video reading class
    Examples:
    ```
        Vcap = VideoCap(mode=0)
        while True:
            img = Vcap.read()
            Vcap.show("ss", img)
    ```
    """
    def __init__(self, mode=0, width=640, height=480, light=150):
        self.cap = cv2.VideoCapture(mode)
        self.cap.set(3, width)
        self.cap.set(4, height)
        self.cap.set(10, light)
        self.start_number = 0

    def read(self, flip=None):
        """
        :param flip: -1: Horizontal and vertical directions,
                      0: along the y-axis, vertical,
                      1: along the x-axis, horizontal
        """
        _, img = self.cap.read()
        if flip is not None:
            assert flip in [-1, 0, 1], f"VideoCap: The 'flip' parameter must be -1, 0, or 1."
            img = cv2.flip(img, flip)
        return img

    def free(self):
        """
        Release camera
        """
        self.cap.release()
        cv2.destroyAllWindows()

    def show(self,
             winname: str,
             src,
             base_name: str = './result.png',
             end_k=27,
             save_k=ord('s'),
             delay_t=1,
             extend_num=3
             ):
        """
        Window display. Press 's' to save, 'Esc' to end
        """
        image_path, ext = os.path.splitext(base_name)
        os.makedirs(os.path.dirname(base_name), exist_ok=True)
        if src is not None:
            cv2.imshow(winname, src)
            k = cv2.waitKey(delay_t) & 0xFF
            if k == end_k:
                self.free()
                sys.exit(0)
            elif k == save_k:
                self.start_number += 1
                file_number = str(self.start_number).zfill(extend_num)
                file_path = f"{image_path}_{file_number}{ext}"
                print(f"{self.start_number}  Image saved to {file_path}")
                cv2.imwrite(file_path, src)

def StackedImagesV1(scale, imgArray):
    """
    Display Images According to List Structure

    :param scale: The scale of the images, where 1 represents the original size.
    :param imgArray: A list of images representing the arrangement in rows and columns.
    :return: A generated image that displays the images in the order specified by the input list, arranged in a grid.
    """
    rows = len(imgArray)
    cols = len(imgArray[0])
    rowsAvailable = isinstance(imgArray[0], list)
    width = imgArray[0][0].shape[1]
    height = imgArray[0][0].shape[0]
    if rowsAvailable:
        for x in range(0, rows):
            for y in range(0, cols):
                if imgArray[x][y].shape[:2] == imgArray[0][0].shape [:2]:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (0, 0), None, scale, scale)
                else:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (imgArray[0][0].shape[1], imgArray[0][0].shape[0]), None, scale, scale)
                if len(imgArray[x][y].shape) == 2: imgArray[x][y]= cv2.cvtColor( imgArray[x][y], cv2.COLOR_GRAY2BGR)
        imageBlank = np.zeros((height, width, 3), np.uint8)
        hor = [imageBlank] * rows
        for x in range(0, rows):
            hor[x] = np.hstack(imgArray[x])
        ver = np.vstack(hor)
    else:
        for x in range(0, rows):
            if imgArray[x].shape[:2] == imgArray[0].shape[:2]:
                imgArray[x] = cv2.resize(imgArray[x], (0, 0), None, scale, scale)
            else:
                imgArray[x] = cv2.resize(imgArray[x], (imgArray[0].shape[1], imgArray[0].shape[0]), None,scale, scale)
            if len(imgArray[x].shape) == 2:
                imgArray[x] = cv2.cvtColor(imgArray[x], cv2.COLOR_GRAY2BGR)
        hor = np.hstack(imgArray)
        ver = hor
    return ver

def StackedImagesV2(scale, imgList, cols):
    """
    Combine multiple images into a single display within a single window
    :param scale: The scaling factor for the images, where a value greater than 1 indicates enlargement and a value less than 1 indicates reduction.
    :param imgList: A list of images to be combined.
    :param cols: The number of images to display per row.
    :return: The combined image.
    """
    totalImages = len(imgList)
    rows = totalImages // cols if totalImages // cols * cols == totalImages else totalImages // cols + 1
    blankImages = cols * rows - totalImages

    width = imgList[0].shape[1]
    height = imgList[0].shape[0]
    imgBlank = np.zeros((height, width, 3), np.uint8)
    imgList.extend([imgBlank] * blankImages)
    for i in range(cols * rows):
        imgList[i] = cv2.resize(imgList[i], (0, 0), None, scale, scale)
        if len(imgList[i].shape) == 2:
            imgList[i] = cv2.cvtColor(imgList[i], cv2.COLOR_GRAY2BGR)
    hor = [imgBlank] * rows
    for y in range(rows):
        line = []
        for x in range(cols):
            line.append(imgList[y * cols + x])
        hor[y] = np.hstack(line)
    ver = np.vstack(hor)
    return ver

def Stackedplt(images, shape):
    """
    Stack multiple images according to the specified shape
    :param images: A list containing multiple PIL Image objects
    :param shape: The number of rows and columns for displaying the images, e.g., (2, 3)
    :return: A stacked PIL Image object
    """
    rows, cols = shape
    new_images = []
    for im in images:
        if is_pil(im):
            new_images.append(im)
        elif is_numpy(im):
            new_images.append(cv2pil(im))
    assert all(img.size == new_images[0].size for img in new_images), "All images must be the same size"
    width, height = new_images[0].size
    new_img_width = width * cols
    new_img_height = height * rows
    new_img = Image.new('RGB', (new_img_width, new_img_height), color=(255, 255, 255))
    x_offset = 0
    y_offset = 0
    for i in range(rows):
        for j in range(cols):
            index = i * cols + j
            if index < len(new_images):
                img = new_images[index]
                new_img.paste(img, (x_offset, y_offset))
                x_offset += width
        x_offset = 0
        y_offset += height

    return new_img

def Stackedtorch(imgs, num_rows, num_cols, titles=None, scale=None, camp=None):
    """
    Display images in a grid format, which can include either Tensor images or PIL images.
    :param imgs: A list containing the images. The list can include both Tensor images and PIL images.
    :param num_rows: The number of rows in the grid.
    :param num_cols: The number of columns in the grid.
    :param titles: A list of titles for the images. Defaults to None.
    :param scale: The scaling factor for the images. Defaults to None, which means no scaling is applied.
    :param camp: 'gray'
    :return:
    Examples:
    ```
        imgs = [image1, image2, image3, image4]
        titles = ['Image 1', 'Image 2', 'Image 3','Image 4']
        axes = Stackedtorch(imgs, 2, 2, titles=titles, scale=5)
        plt.savefig('my_plot.jpg')
        plt.show()
    ```
    """
    if scale:
        figsize = (num_cols * scale, num_rows * scale)
        _, axes = plt.subplots(num_rows, num_cols, figsize=figsize)
    else:
        _, axes = plt.subplots(num_rows, num_cols)
    axes = axes.flatten()
    for i, (ax, img) in enumerate(zip(axes, imgs)):
        if is_tensor(img):
            # Tensor Image
            ax.imshow(img.numpy(), cmap=camp)
        else:
            # PIL Image
            ax.imshow(img, cmap=camp)
        ax.axes.get_xaxis().set_visible(False)
        ax.axes.get_yaxis().set_visible(False)
        if titles:
            ax.set_title(titles[i])

    return axes

def Mp4toGif(mp4, name='result.gif', fps=10, start=None, end=None):
    """Convert MP4 files to GIF animations"""
    cap = cv2.VideoCapture(mp4)
    all_images = []
    frame_count = 0
    while True:
        ret, img = cap.read()
        if ret is False:
            break
        if start is not None and frame_count < start:
            frame_count += 1
            continue
        if end is not None and frame_count >= end:
            break
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        all_images.append(img)
        frame_count += 1

    duration = int(1000 / fps)  # 将帧率转换为每帧之间的延迟时间（毫秒）
    imageio.mimsave(name, all_images, duration=duration)
    print("Conversion completed！")


"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is used for image drawing.
"""
import cv2
import numpy as np
import pyzjr.Z as Z

__all__ = ["OverlayPng", "PutBoxText", "AddText", "ConvertBbox", "DrawPolygon", "CornerRect"]

def OverlayPng(imgBack, imgFront, pos=(0, 0), alpha_gain=1.0):
    """
    Overlay display image with proper alpha blending
    :param imgBack: Background image, no format requirement, 3 channels
    :param imgFront: PNG pre image, must be read using cv2.IMREAD_UNCHANGED=-1
    :param pos: Placement position

    Examples:
    '''
        background = cv2.imread(background_path)
        overlay = cv2.imread(overlay_path, cv2.IMREAD_UNCHANGED)
        fused_image = pyzjr.OverlayPng(background, overlay, alpha_gain=1.5)
    '''
    """
    hf, wf, cf = imgFront.shape
    hb, wb, cb = imgBack.shape
    y_pos, x_pos = pos
    y_end = y_pos + hf
    x_end = x_pos + wf

    # Ensure we don't go beyond the background boundaries
    if y_end > hb:
        y_end = hb
    if x_end > wb:
        x_end = wb

    # Resize overlay to fit the background (optional but good practice)
    overlay_resized = cv2.resize(imgFront, (x_end - x_pos, y_end - y_pos))

    overlay_alpha = overlay_resized[:, :, 3].astype(float) / 255.0
    overlay_alpha = np.clip(overlay_alpha * alpha_gain, 0, 1)
    background_alpha = 1.0 - overlay_alpha

    result = overlay_resized[:, :, :3] * overlay_alpha[..., np.newaxis] + imgBack[y_pos:y_end, x_pos:x_end, :3] * background_alpha[..., np.newaxis]

    imgBack[y_pos:y_end, x_pos:x_end, :3] = result.astype(np.uint8)

    return imgBack


def PutBoxText(imgBack, bbox, text, mode=0, bboxcolor=Z.green, textcolor=Z.red,
               fontsize=.5, thickness=2, bboxthickness=2, font=cv2.FONT_HERSHEY_SIMPLEX):
    """
    Add text within a bounding box on an image.

    Args:
        imgBack (numpy.ndarray): The background image where the text and bounding box will be drawn.
        bbox (list of int): The bounding box coordinates in the format [x1, y1, x2, y2].
        text (str): The text to be drawn.
        mode (int): Determines the position of the text within the bounding box.
            - -1: Top-left corner.
            - 0: Centered.
            - 1: Top-right corner.
        bboxcolor: The RGB color of the bounding box.
        textcolor: The RGB color of the text.
        fontsize (float): Font size of the text.
        thickness (int): Thickness of the text.
        bboxthickness (int): Thickness of the bounding box.
        font (int): Font type to be used.
    """
    text_x = text_y = None
    x1, y1, x2, y2 = bbox
    text_size, _ = cv2.getTextSize(text, font, fontsize, thickness)
    cv2.rectangle(imgBack, (x1, y1), (x2, y2), bboxcolor, bboxthickness)

    if mode == 0:
        # Centered
        text_x = int((x1 + x2) / 2 - text_size[0] / 2)
        text_y = int((y1 + y2) / 2 + text_size[1] / 2)
    elif mode == -1:
        # Top-left corner
        text_x = int(x1)
        text_y = int(y1 - text_size[1])
    elif mode == 1:
        # Top-right corner
        text_x = int(x2 - text_size[0])
        text_y = int(y1 - text_size[1])

    cv2.putText(imgBack, text, (text_x, text_y), font, fontsize, textcolor, thickness)


def AddText(img, text, x, y, color, thickness=3):
    """
    Usage Directions:
        Using the OpenCV method to add text to images
    Code Example:
    ------------
    >>> img = np.zeros((200, 400, 3), dtype=np.uint8)
    >>> AddText(img, "Hello, OpenCV!", 50, 100, (0, 255, 0), 2)
    """
    cv2.putText(
        img,
        text,
        (x, y),
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=1,
        thickness=thickness,
        color=color,
        lineType=cv2.LINE_AA,
    )


def DrawPolygon(points, image, color):
    """
    Usage Directions:
        Draw a polygon onto an image using the given points and fill color.
    Code Example:
    ------------
    >>> image = np.ones((400, 400, 3), dtype=np.uint8)
    >>> polygon_points = np.array([[100, 100], [200, 50], [300, 150], [250, 300], [150, 250]])
    >>> result_image = DrawPolygon(polygon_points, image, (255, 0, 0))
    """
    points = np.array([points])
    points = points.astype(np.int32)
    image = cv2.fillPoly(image, points, color)
    return image


def ConvertBbox(bbox, to_center=True):
    """
    Convert bounding box coordinates between (upper left, lower right) and (center, width, height).

    :param bbox: List of bounding boxes in either [x1, y1, x2, y2] or [cx, cy, w, h] format.
    :param to_center: If True, convert from (upper left, lower right) to (center, width, height).
                      If False, convert from (center, width, height) to (upper left, lower right).
    :return: List of converted bounding boxes.
    """
    converted_boxes = []
    for box in bbox:
        if to_center:
            # Convert from (upper left, lower right) to (center, width, height)
            x1, y1, x2, y2 = box
            cx = (x1 + x2) / 2
            cy = (y1 + y2) / 2
            w = x2 - x1
            h = y2 - y1
            converted_box = [cx, cy, w, h]
        else:
            # Convert from (center, width, height) to (upper left, lower right)
            cx, cy, w, h = box
            x1 = cx - 0.5 * w
            y1 = cy - 0.5 * h
            x2 = cx + 0.5 * w
            y2 = cy + 0.5 * h
            converted_box = [x1, y1, x2, y2]
        converted_boxes.append(converted_box)
    return converted_boxes


def CornerRect(drawimg, bbox, length=30, lthickness=5, rthickness=1,
               bboxcolor=Z.purple, colorCorner=Z.green):
    """
    Draw a rectangle with corner marks on an image
    :param drawimg: The image to draw on
    :param bbox: The bounding box, represented as [x1, y1, x2, y2]
    :param length: The length of the corner marks
    :param lthickness: The line thickness of the corner marks
    :param rthickness: The line thickness of the rectangle
    :param bboxcolor: The color of the rectangle
    :param colorCorner: The color of the corner marks
    :return: A copy of the image with corner marks drawn
    """
    x, y, x1, y1 = bbox
    cv2.rectangle(drawimg, (x, y), (x1, y1), bboxcolor, rthickness)
    # Top-left corner (x, y)
    cv2.line(drawimg, (x, y), (x + length, y), colorCorner, lthickness)
    cv2.line(drawimg, (x, y), (x, y + length), colorCorner, lthickness)
    # Top-right corner (x1, y)
    cv2.line(drawimg, (x1, y), (x1 - length, y), colorCorner, lthickness)
    cv2.line(drawimg, (x1, y), (x1, y + length), colorCorner, lthickness)
    # Bottom-left corner (x, y1)
    cv2.line(drawimg, (x, y1), (x + length, y1), colorCorner, lthickness)
    cv2.line(drawimg, (x, y1), (x, y1 - length), colorCorner, lthickness)
    # Bottom-right corner (x1, y1)
    cv2.line(drawimg, (x1, y1), (x1 - length, y1), colorCorner, lthickness)
    cv2.line(drawimg, (x1, y1), (x1, y1 - length), colorCorner, lthickness)

if __name__=="__main__":
    import cv2
    imagePath = r"D:\PythonProject\pyzjr\pyzjr\test.png"
    image = np.ones((400, 400, 3), dtype=np.uint8)

    img = cv2.imread(imagePath)
    CornerRect(img, [50, 50, 200, 200])
    # PutBoxText(img, [50, 50, 200, 200], text="DOG", mode=0)
    cv2.imshow("show", img)
    cv2.waitKey(0)



# Skeleton extraction
"""
skeletoncv、sketionio均是使用zhang-suen骨架算法（手写算法不佳），sketion_medial_axis采用的是中轴提取
下面的膨胀、中值滤波（滤波效果中最佳）、闭运算是经过实验进行的组合，针对自己的数据集可以先尝试如果效果不佳再采取实验
"""
import cv2
import numpy as np
import skimage.filters as filters
from skimage.morphology import skeletonize, dilation, disk
from skimage import io, morphology
from pyzjr.augmentation.mask_ops import BinaryImg, medial_axis_mask

def skeletoncv(single_pic_path):
    image = cv2.imread(single_pic_path)
    binary = BinaryImg(image)
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
    binary = cv2.dilate(binary, kernel, iterations=1)
    binary = cv2.medianBlur(binary, 5)
    binary = cv2.dilate(binary, kernel, iterations=1)
    binary = cv2.medianBlur(binary, 5)
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
    binary = cv2.morphologyEx(binary, cv2.MORPH_CLOSE, kernel)
    skeleton = skeletonize(binary / 255)
    skeleton = (skeleton * 255).astype(np.uint8)
    return skeleton

def sketionio(single_pic_path):
    image = io.imread(single_pic_path, as_gray=True)
    thresh = filters.threshold_otsu(image)
    binary = image > thresh

    binary = dilation(binary, disk(3))
    binary = filters.median(binary, footprint=morphology.disk(5))
    binary = dilation(binary, disk(2))
    binary = filters.median(binary, footprint=morphology.disk(5))

    selem = morphology.disk(3)
    binary = morphology.closing(binary, selem)

    skeleton = skeletonize(binary)
    skeleton = skeleton.astype(np.uint8) * 255
    return skeleton

def sketion_medial_axis(single_pic_path):
    image = cv2.imread(single_pic_path)
    binary = BinaryImg(image)

    binary = dilation(binary, disk(2))
    binary = filters.median(binary, footprint=morphology.disk(5))
    binary = dilation(binary, disk(2))
    binary = filters.median(binary, footprint=morphology.disk(5))
    selem = morphology.disk(3)
    binary = morphology.closing(binary, selem)
    binary = medial_axis_mask(binary)
    return binary.astype(np.uint8)


if __name__=="__main__":
    from pyzjr.augmentation.mask_ops import unique
    from pyzjr.augmentation.mask_ops import count_nonzero
    path = r'D:\PythonProject\pyzjrPyPi\models_img\1604.png'
    image_cv = skeletoncv(path)
    print(unique(image_cv))
    print(image_cv.shape)
    print(count_nonzero(image_cv))

    image_io = sketionio(path)
    print(unique(image_io))
    print(image_io.shape)

    image_axis = sketion_medial_axis(path)
    print(unique(image_axis))
    print(image_axis.shape)

    cv2.imshow("test skeleton function cv", image_cv)
    cv2.imshow("test skeleton function io", image_io)
    cv2.imshow("test skeleton function medial axis", image_axis)
    cv2.waitKey(0)


from .clean import clear_directory, cleanup_directory

from .folder import (
    multi_makedirs, unique_makedirs, datatime_makedirs, logdir, loss_weights_dirs, rm_makedirs
)
from .file import (
    file_age,
    file_date,
    file_size,
    read_yaml,
    read_json
)
from .txtfun import generate_txt, read_txt, write_to_txt
from .getpath import (
    getPhotoPath,
    SearchFilePath,
    SearchFileName,
    split_path2list,
    AbsPathOps
)


"""
BAM: Bottleneck Attention Module
Original paper addresshttps: https://arxiv.org/pdf/1807.06514.pdf
The code comes from the official implementation:
https://github.com/Jongchan/attention-module/tree/master
"""
import torch
import torch.nn as nn
import torch.nn.functional as F

class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.size(0), -1)

class ChannelGate(nn.Module):
    def __init__(self, gate_channel, reduction_ratio=16, num_layers=1):
        super(ChannelGate, self).__init__()
        self.gate_c = nn.Sequential()
        self.gate_c.add_module('flatten', Flatten())
        gate_channels = [gate_channel]
        gate_channels += [gate_channel // reduction_ratio] * num_layers
        gate_channels += [gate_channel]
        for i in range(len(gate_channels) - 2):
            self.gate_c.add_module('gate_c_fc_%d' % i, nn.Linear(gate_channels[i], gate_channels[i+1]))
            self.gate_c.add_module('gate_c_bn_%d' % (i+1), nn.BatchNorm1d(gate_channels[i+1]))
            self.gate_c.add_module('gate_c_relu_%d' % (i+1), nn.ReLU())
        self.gate_c.add_module('gate_c_fc_final', nn.Linear(gate_channels[-2], gate_channels[-1]) )

    def forward(self, in_tensor):
        avg_pool = F.avg_pool2d(in_tensor, in_tensor.size(2), stride=in_tensor.size(2))
        return self.gate_c(avg_pool).unsqueeze(2).unsqueeze(3).expand_as(in_tensor)

class SpatialGate(nn.Module):
    def __init__(self, gate_channel, reduction_ratio=16, dilation_conv_num=2, dilation_val=4):
        super(SpatialGate, self).__init__()
        self.gate_s = nn.Sequential()
        self.gate_s.add_module('gate_s_conv_reduce0', nn.Conv2d(gate_channel, gate_channel//reduction_ratio, kernel_size=1))
        self.gate_s.add_module('gate_s_bn_reduce0',	nn.BatchNorm2d(gate_channel//reduction_ratio) )
        self.gate_s.add_module('gate_s_relu_reduce0', nn.ReLU(inplace=True))
        for i in range(dilation_conv_num):
            self.gate_s.add_module('gate_s_conv_di_%d' % i, nn.Conv2d(gate_channel//reduction_ratio, gate_channel//reduction_ratio,
                                                                       kernel_size=3, padding=dilation_val, dilation=dilation_val))
            self.gate_s.add_module('gate_s_bn_di_%d' % i, nn.BatchNorm2d(gate_channel//reduction_ratio))
            self.gate_s.add_module('gate_s_relu_di_%d' % i, nn.ReLU(inplace=True))
        self.gate_s.add_module( 'gate_s_conv_final', nn.Conv2d(gate_channel//reduction_ratio, 1, kernel_size=1))

    def forward(self, in_tensor):
        return self.gate_s(in_tensor).expand_as(in_tensor)

class BAMAttention(nn.Module):
    def __init__(self, gate_channel):
        super(BAMAttention, self).__init__()
        self.channel_att = ChannelGate(gate_channel)
        self.spatial_att = SpatialGate(gate_channel)

    def forward(self,x):
        att = 1 + torch.sigmoid(self.channel_att(x) * self.spatial_att(x))
        return att * x

if __name__ == "__main__":
    input_tensor = torch.randn((2, 64, 32, 32))
    block = BAMAttention(64)
    output_tensor = block(input_tensor)
    print(output_tensor.shape)


"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is used to check the input type.
"""
import re
import os
import torch
import torch.nn as nn
import numpy as np
from PIL import Image
import pyzjr.Z as Z
import platform
from typing import Iterable

def is_tensor(img):
    """Check if the input image is torch format."""
    return isinstance(img, torch.Tensor)

def is_pil(img):
    """Check if the input image is PIL format."""
    return isinstance(img, Image.Image)

def is_numpy(img):
    """Check if the input image is Numpy format."""
    return isinstance(img, np.ndarray)

def is_gray_image(image):
    return (len(image.shape) == 2) or (len(image.shape) == 3 and image.shape[-1] == 1)

def is_rgb_image(image):
    return len(image.shape) == 3 and image.shape[-1] == 3

def is_parallel(model):
    # Returns True if model is of type DP or DDP
    return isinstance(model, (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel))

def is_Iterable(param):
    return isinstance(param, Iterable)

def is_str(param):
    return isinstance(param, str)

def is_int(param):
    return isinstance(param, int)

def is_float(param):
    return isinstance(param, float)

def is_bool(param):
    return param.dtype == np.bool_

def is_list(param):
    return isinstance(param, list)

def is_tuple(param):
    return isinstance(param, tuple)

def is_list_or_tuple(param):
    return isinstance(param, (list, tuple))

def is_none(param):
    return True if param is None else False

def is_not_none(param):
    return not is_none(param)

def is_positive_int(param):
    return is_int(param) and param > 0

def is_nonnegative_int(param):
    return is_int(param) and param >= 0

def is_ascii(s):
    """Check if the string is composed of only ASCII characters."""
    s = str(s)
    return all(ord(c) < 128 for c in s)

def is_url(filename):
    """Return True if string is an http or ftp path."""
    URL_REGEX = re.compile(r'http://|https://|ftp://|file://|file:\\')
    return (isinstance(filename, str) and
            URL_REGEX.match(filename) is not None)

def is_image_extension(image_name):
    ext = image_name.split('.')
    return ext[-1] in Z.IMG_FORMATS and is_str(image_name)

def is_video_extension(video_name):
    ext = video_name.split('.')
    return ext[-1] in Z.VID_FORMATS and is_str(video_name)

def is_file(path):
    return os.path.isfile(path) and is_str(path)

def is_directory(path):
    return os.path.isdir(path) and is_str(path)

def is_directory_not_empty(path):
    return os.path.isdir(path) and len(os.listdir(path)) > 0 and is_str(path)

def is_path_exists(path):
    return os.path.exists(path) and is_str(path)

def is_windows():
    if platform.system() == "Windows":
        return True

def is_linux():
    if platform.system() == "Linux":
        return True

def is_odd(n):
    return n % 2 != 0

def is_even(n):
    return n % 2 == 0

def _check_input_is_cv(img):
    if not is_numpy(img):
        raise TypeError("The loaded image should conform to the format read by OpenCV!")

def _check_input_is_plt(img):
    if not is_pil(img):
        raise TypeError("The loaded image should conform to the format read by PIL!")

def _check_input_is_tensor(img):
    if not is_tensor(img):
        raise TypeError('Input tensor should be a torch tensor. Got {}.'.format(type(img)))

def get_image_size(image):
    if is_numpy(image):
        h, w = image.shape[:2]
        return h, w

    if is_pil(image):
        w, h = image.size
        return h, w

    if is_tensor(image):
        if len(image.shape) == 4 or len(image.shape) == 3:
            w, h = image.shape[-2:]
            return h, w
    else:
        raise ValueError("Unsupported input type")

def get_image_num_channels(img):
    if is_tensor(img):
        if img.ndim == 2:
            return 1
        elif img.ndim > 2:
            return img.shape[-3]

    if is_pil(img):
        return 1 if img.mode == 'L' else 3

    if is_numpy(img):
        return 1 if is_gray_image(img) else 3

if __name__=="__main__":
    image_name = r'D:\PythonProject\pyzjrPyPi\pyzjr\augmentation\test.png'
    print(is_file(image_name), is_windows())


"""
Copyright (c) 2023, Auorui.
All rights reserved.
"""
import cv2
import numpy as np
from PIL import Image
from pyzjr.utils.check import is_numpy, is_bool
from skimage.morphology import medial_axis
from scipy.ndimage import distance_transform_edt

def convert_np(image, dtype=np.uint8):
    """Convert an image to a NumPy array with the specified dtype."""
    return image if is_numpy(image) else np.array(image).astype(dtype)

def clip(image, maxval, dtype=np.uint8):
    """
    Truncate the pixel values of the image to the specified range and perform data type conversion
    """
    np_image = convert_np(image)
    return np.clip(np_image, 0, maxval).astype(dtype)

def unique(image):
    """Returns a list of unique pixel values in the input image"""
    np_image = convert_np(image)
    return np.unique(np_image)

def RectMask(image, up_left_coord=None, lower_right_coord=None, bboxes=None):
    """
    Create a rectangular mask
    :param image: Original image.
    :param up_left_coord: [x1, y1]
    :param lower_right_coord: [x2, y2]
    :param bboxes: [x1, y1, x2, y2]
    """
    image = convert_np(image)
    h, w = image.shape[:2]
    mask = np.zeros((h, w, 3), dtype=np.uint8)
    if bboxes is None:
        x1, y1 = up_left_coord
        x2, y2 = lower_right_coord
    else:
        x1, y1, x2, y2 = bboxes
    mask[y1: y2, x1: x2] = 255

    if mask.ndim == 3 and mask.shape[-1] > 1:
        mask = mask[:, :, 0]
    masked_image = cv2.bitwise_and(image, image, mask=mask)
    return mask, masked_image

def BinaryImg(image, min_value=127, max_value=255):
    """
    Convert image to binary image
    :param image: Original image. --- (BGR format)
    :param min_value: Minimum threshold, default to 127
    :param max_value: Maximum threshold, default to 225
    """
    np_image = convert_np(image)
    gray_image = cv2.cvtColor(np_image, cv2.COLOR_BGR2GRAY)
    _, binary_image = cv2.threshold(gray_image, min_value, max_value, cv2.THRESH_BINARY)
    return binary_image

def up_low(image, lower, upper, dtype=np.uint8):
    """
    Create a binarized mask based on the given color range (lower and upper limits)
    :param lower: Lower limit of color range, (B, G, R)
    :param upper: Upper limit of color range, (B, G, R)
    """
    np_image = convert_np(image)
    lower = np.array(lower, dtype=dtype)
    upper = np.array(upper, dtype=dtype)
    _mask = cv2.inRange(np_image, lower, upper)
    return _mask

def approximate(image, std=127.5, dtype=np.uint8):
    """
    Convert a single channel image into a binary image.
    """
    image[image > std] = 255
    image[image < std] = 0
    image = image.astype(dtype)
    return image

def ceilfloor(image, dtype=np.uint8):
    """
    The pixel value of the input image is limited between the maximum value of 255 and the minimum value of 0
    """
    image[image > 255] = 255
    image[image < 0] = 0
    image = image.astype(dtype)
    return image

def bool2mask(matrix):
    """
    Convert Boolean matrix to 8-bit unsigned integer mask image
    """
    if is_bool(matrix):
        result_int = matrix.astype(int)
        _mask = result_int * 255
        return _mask.astype(np.uint8)
    else:
        raise ValueError("Input matrix must be of bool dtype.")

def cv_distance(image):
    """Using distance transformation functions in OpenCV"""
    dist_transform = cv2.distanceTransform(image, cv2.DIST_L2, 3)
    return dist_transform

def chamfer(image, weight=None):
    """Chamfer distance transformation"""
    if weight is None:
        weights = np.array([[1, 1, 1],
                            [1, 0, 1],
                            [1, 1, 1]], dtype=np.uint8)
    else:
        weights = weight
    dist_transform = cv2.filter2D(image, cv2.CV_32F, weights)
    return dist_transform

def fast_marching(image):
    """Fast-Marching Distance Transform"""
    _, medial_axis_image = medial_axis(image, return_distance=True)   # 使用medial_axis函数计算中轴线
    dist_transform = distance_transform_edt(medial_axis_image)
    return dist_transform

def inpaint_defect(image, mask, radius=10, flags=1):
    """
    Inpaint defect image
    :param image: defect image
    :param mask: Mask of defective parts
    :param radius: Inpaint radius
    :param flags: TELEA-1; NS-0
    """
    dst = cv2.inpaint(image, mask, radius, flags)
    return dst

def medial_axis_mask(image):
    """
    The input must be a binary graph to obtain the axis image within it
    """
    result = medial_axis(image)
    return bool2mask(result)

def cvt8png(pngpath, bit_depth=False, target=(255, 255, 255), convert=(128, 0, 0)):
    """
    Voc: RGB png彩图转换
    :param bit_depth: 默认转为8位,需要使用out_png.save可以正确保存
    True:                      False:
        plt.imshow(img)             cv2.imshow("img",img)
        plt.axis('off')             cv2.waitKey(0)
        plt.show()
    :param pngpath: 为保证是按照cv2的方式读入, 所以传入路径即可
    :param target: 目标颜色, RGB方式
    :param convert: 转换颜色, 同RGB格式
    :return:
    """
    png = cv2.imread(pngpath)
    png = cv2.cvtColor(png, cv2.COLOR_BGR2RGB)
    h, w = png.shape[:2]

    mask = np.all(png == target, axis=-1)
    out_png = np.zeros((h, w, 3), dtype=np.uint8)
    out_png[mask] = convert
    out_png[~mask] = png[~mask]

    if bit_depth:
        out_png_pil = Image.fromarray(out_png)
        out_png_pil = out_png_pil.convert("P", palette=Image.ADAPTIVE, colors=256)
        return out_png_pil
    else:
        out_png_cv = cv2.cvtColor(out_png, cv2.COLOR_RGB2BGR)
        return out_png_cv

def cvtMask(pngpath, lower ,upper, convert=(255, 255, 255)):
    """
    去除标签中的杂色,并对其进行颜色的转换

    :param pngpath: 输入图像的文件路径
    :param lower: 颜色范围的下限值，作为一个包含(B, G, R)值的元组
    :param upper: 颜色范围的上限值，作为一个包含(B, G, R)值的元组
    :param convert: 需要转换为的颜色，作为一个包含(B, G, R)值的元组
    :return: 处理后的图像
    """
    image = cv2.imread(pngpath)
    lower = np.array(lower, dtype=np.uint8)
    upper = np.array(upper, dtype=np.uint8)

    mask = cv2.inRange(image, lower, upper)
    image[mask > 0] = convert

    return image

def imageContentMove(image, H_pixels=None, V_pixels=None):
    if H_pixels is None:
        H_pixels = 0
    if V_pixels is None:
        V_pixels = 0
    h, w = image.shape[:2]
    black = np.zeros_like(image)

    start_row, end_row = max(0, V_pixels), min(h, h + V_pixels)
    start_col, end_col = max(0, H_pixels), min(w, w + H_pixels)

    fill_start_row, fill_end_row = max(0, -V_pixels), min(h, h - V_pixels)
    fill_start_col, fill_end_col = max(0, -H_pixels), min(w, w - H_pixels)

    black[start_row:end_row, start_col:end_col] = image[fill_start_row:fill_end_row, fill_start_col:fill_end_col]

    return black

def count_zero(thresh):
    """计算矩阵0值"""
    zero_count = np.sum(thresh == 0)
    return zero_count

def count_white(thresh):
    """计算矩阵255值"""
    white_count = np.sum(thresh == 255)
    return white_count

def count_nonzero(thresh):
    """计算矩阵非0值"""
    nonzero_pixels = np.count_nonzero(thresh)
    return nonzero_pixels

if __name__=="__main__":
    image = np.zeros((10, 10), dtype=np.uint8)
    image[2:8, 2:8] = 255
    print(image)

    dist_transform = cv_distance(image)
    chamfer_dist_transform = chamfer(image)
    fast_marching_dist_transform = fast_marching(image)
    # 距离变换结果
    print("Distance Transform:")
    print(dist_transform)
    print("\nChamfer Distance Transform:")
    print(chamfer_dist_transform)
    print("\nFast-Marching Distance Transform:")
    print(fast_marching_dist_transform)

    mask_img_path = r'D:\PythonProject\pyzjrPyPi\models_img\1604.png'
    mask_img = cv2.imread(mask_img_path)
    img = clip(mask_img, 255)
    # img = up_low(img, (0, 0, 100), (255, 255, 255))
    # img = approximate(img, 130)
    img = BinaryImg(img)
    # img = ceilfloor(img)
    # _, img = RectMask(img, (15, 27), (388, 300))
    result = medial_axis_mask(img)
    print(unique(result))
    # print(result.shape)
    cv2.imshow("test mask-function", result)
    cv2.waitKey(0)




from .OneHot import (
    one_hot,
    get_one_hot,
    get_one_hot_with_np,
    get_one_hot_with_torch,
)
from .learnrate import (
    get_optimizer,
    CustomScheduler,
    FixedStepLR,
    MultiStepLR,
    CosineAnnealingLR,
    WarmUpLR,
    WarmUp,
    FindLR,
    lr_finder
)
from .avgweight import (
    AveragingBaseModel,
    EMAModel,
    SWAModel,
    T_ADEMAModel,
    de_parallel,
    get_ema_avg_fn,
    get_swa_avg_fn,
    get_t_adema_fn
)
from .lr_scheduler import _LRScheduler
from .functional import *


import os
from pathlib import Path

import pyzjr.Z as Z
from pyzjr.data.base import natsorted

def getPhotoPath(path, debug=False):
    """
    :param path: 文件夹路径
    :param debug: 开启打印文件名错误的名字
    :return: 包含图片路径的列表
    """
    imgfile = []
    allfile = []
    file_list = os.listdir(path)
    for i in file_list:
        if debug:
            if i[0] in ['n', 't', 'r', 'b', 'f'] or i[0].isdigit():
                print(f"File name error occurred at the beginning of {i}!")
        newph = os.path.join(path, i).replace("\\", "/")
        allfile.append(newph)
        _, file_ext = os.path.splitext(newph)
        if file_ext[1:] in Z.IMG_FORMATS:
            imgfile.append(newph)

    return natsorted(imgfile), natsorted(allfile)

def SearchFilePath(filedir, file_ext='.png'):
    """What is returned is a list that includes all paths under the target path that match the suffix."""
    search_file_path = []
    for root, dirs, files in os.walk(filedir):
        for filespath in files:
            if str(filespath).endswith(file_ext):
                search_file_path.append(os.path.join(root, filespath))
    return natsorted(search_file_path)

def SearchFileName(target_path, file_ext='.png'):
    """Only search for file names in the appropriate format in the target folder"""
    all_files = os.listdir(target_path)
    png_files = [file for file in all_files if file.lower().endswith(file_ext)]
    sorted_png_files = natsorted(png_files)
    return sorted_png_files

def split_path2list(path_str):
    """
    path_list = split_path2list('D:\PythonProject\MB_TaylorFormer\DehazeFormer\data\rshazy\test\GT\220.png')
    Return:
        ['D:\\', 'PythonProject', 'MB_TaylorFormer', 'DehazeFormer', 'data', 'rshazy', 'test', 'GT', '220.png']
    """
    path = Path(path_str)
    path_parts = path.parts
    return list(path_parts)

class AbsPathOps():
    def __init__(self, path):
        self.path_list = split_path2list(path)
        self._full_path = path

    @property
    def basename(self):
        return self.path_list[-1]

    @property
    def format(self):
        basename = self.basename
        if "." in basename:
            return basename.split(".")[-1]
        else:
            return ""

    @property
    def disk(self):
        if not os.path.isabs(path):
            raise ValueError(f"The provided path {path} is not an absolute path.")
        return self.path_list[0]

    @property
    def dirname(self):
        return os.sep.join(self.path_list[:-1])

    @property
    def filename(self):
        basename = self.basename
        if "." in basename:
            return basename.rsplit(".", 1)[0]  # 使用 rsplit 从右侧开始分割，并限制为分割一次
        else:
            return basename

    @property
    def parent(self):
        return os.path.dirname(self.dirname)


if __name__ == "__main__":
    path = r"D:\PythonProject\pyzjrPyPi\pyzjr\data"
    print(SearchFilePath(path))
    pathops = AbsPathOps(path)
    print(pathops.basename)
    print(pathops.format)
    print(pathops.disk)
    print(pathops.dirname)
    print(pathops.filename)
    print(pathops.parent)


#: Loss binary mode suppose you are solving binary segmentation task.
#: That mean yor have only one class which pixels are labled as **1**,
#: the rest pixels are background and labeled as **0**.
#: Target mask shape - (N, H, W), model output mask shape (N, 1, H, W).
BINARY_MODE: str = "binary"

#: Loss multiclass mode suppose you are solving multi-**class** segmentation task.
#: That mean you have *C = 1..N* classes which have unique label values,
#: classes are mutually exclusive and all pixels are labeled with theese values.
#: Target mask shape - (N, H, W), model output mask shape (N, C, H, W).
MULTICLASS_MODE: str = "multiclass"

#: Loss multilabel mode suppose you are solving multi-**label** segmentation task.
#: That mean you have *C = 1..N* classes which pixels are labeled as **1**,
#: classes are not mutually exclusive and each class have its own *channel*,
#: pixels in each channel which are not belong to class labeled as **0**.
#: Target mask shape - (N, C, H, W), model output mask shape (N, C, H, W).
MULTILABEL_MODE: str = "multilabel"



"""
Copyright (c) 2025, Auorui.
All rights reserved.

AN IMAGE IS WORTH 16X16 WORDS: TRANSFORMERS FOR IMAGE RECOGNITION AT SCALE
    <https://arxiv.org/pdf/2010.11929.pdf>
Blog records: https://blog.csdn.net/m0_62919535/article/details/144936876
"""
from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F
from pyzjr.utils.converter import to_2tuple
from pyzjr.nn.models.bricks.drop import DropPath

LayerNorm = partial(nn.LayerNorm, eps=1e-6)

class PatchEmbedding(nn.Module):
    def __init__(self, img_size=224, patch_size=16, in_channels=3, embed_dim=768, norm_layer=None):
        super().__init__()
        self.img_size = to_2tuple(img_size)
        self.patch_size = to_2tuple(patch_size)
        self.embed_dim = embed_dim
        self.norm = norm_layer(self.embed_dim) if norm_layer else nn.Identity()
        self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=patch_size, stride=patch_size, padding=0)

    def forward(self, x):
        x = self.proj(x) # 结果形状为 (batch_size, embed_dim, num_patches_H, num_patches_W)
        x = x.flatten(2) # 将输出展平成 (batch_size, embed_dim, num_patches)
        x = x.transpose(1, 2) # 转置为 (batch_size, num_patches, embed_dim)
        x = self.norm(x)
        return x

class MultiheadAttention(nn.Module):
    def __init__(
            self,
            embed_dim,
            num_heads=8,
            qkv_bias=False,
            attention_dropout_ratio=0.,
            proj_drop=0.,
    ):
        super(MultiheadAttention, self).__init__()
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads
        self.scale = self.head_dim ** -0.5
        self.qkv = nn.Linear(embed_dim, embed_dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attention_dropout_ratio)
        self.out_linear = nn.Linear(embed_dim, embed_dim)
        self.out_linear_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[:3]
        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = F.softmax(attn, dim=-1)
        attn = self.attn_drop(attn)
        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.out_linear(x)
        x = self.out_linear_drop(x)
        return x

class MLP(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop_ratio=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop_ratio)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x

class EncoderBlock(nn.Module):
    """Transformer encoder block.
    在 mlp block中, MLP 层的隐藏维度是输入的维度的4倍,
    详见 Table 1: Details of Vision Transformer model variants
    """
    mlp_ratio = 4
    def __init__(
            self,
            dim,
            num_heads,
            qkv_bias=False,
            drop_ratio=0.,
            attention_dropout_ratio=0.,
            drop_path_ratio=0.,
            norm_layer=LayerNorm,
            act_layer=nn.GELU
    ):
        super(EncoderBlock, self).__init__()
        self.num_heads = num_heads
        # Attention block
        self.norm1 = norm_layer(dim)
        self.attention = MultiheadAttention(dim, num_heads, qkv_bias=qkv_bias, attention_dropout_ratio=attention_dropout_ratio, proj_drop=drop_ratio)
        self.drop_path = DropPath(drop_path_ratio) if drop_path_ratio > 0. else nn.Identity()
        # MLP block
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * self.mlp_ratio)
        self.mlp = MLP(dim, mlp_hidden_dim, drop_ratio=drop_ratio, act_layer=act_layer)

    def forward(self, x):
        x = x + self.drop_path(self.attention(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class TransformerEncoder(nn.Module):
    """堆叠 L 次 Transformer encoder block"""
    def __init__(
            self,
            num_layers,
            dim,
            num_heads,
            qkv_bias=False,
            drop_ratio=0.,
            attention_dropout_ratio=0.,
            drop_path_ratio=0.,
            norm_layer=LayerNorm,
            act_layer=nn.GELU
    ):
        super(TransformerEncoder, self).__init__()
        dpr = [x.item() for x in torch.linspace(0, drop_path_ratio, num_layers)]  # stochastic depth decay rule
        self.layers = nn.ModuleList([
            EncoderBlock(
                dim=dim,
                num_heads=num_heads,
                qkv_bias=qkv_bias,
                drop_ratio=drop_ratio,
                attention_dropout_ratio=attention_dropout_ratio,
                drop_path_ratio=dpr[_],
                norm_layer=norm_layer,
                act_layer=act_layer
            )
            for _ in range(num_layers)
        ])
        self.norm = norm_layer(dim)

    def forward(self, x):
        for layer in self.layers:
            x = layer(x)
        x = self.norm(x)
        return x

class VisionTransformer(nn.Module):
    def __init__(
            self,
            img_size=224,
            patch_size=16,
            in_channels=3,
            num_classes=1000,
            hidden_dim=768,
            num_heads=12,
            num_layers=12,
            qkv_bias=True,
            drop_ratio=0.,
            attention_dropout_ratio=0.,
            drop_path_ratio=0.,
            norm_layer=LayerNorm,
            act_layer=nn.GELU
    ):
        super(VisionTransformer, self).__init__()
        assert img_size == 224, f"Image size must be 224, but got {img_size}"
        assert img_size % patch_size == 0, f"Image size {img_size} must be divisible by patch size {patch_size}"
        self.num_classes = num_classes
        self.num_tokens = 1
        self.patch_embed = PatchEmbedding(img_size=img_size, patch_size=patch_size, in_channels=in_channels,
                                          embed_dim=hidden_dim, norm_layer=norm_layer)
        num_patches = (img_size // patch_size) * (img_size // patch_size)
        self.cls_token = nn.Parameter(torch.zeros(1, 1, hidden_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + self.num_tokens, hidden_dim))
        self.pos_drop = nn.Dropout(p=drop_ratio)

        self.blocks = TransformerEncoder(
            num_layers=num_layers,
            dim=hidden_dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            drop_ratio=drop_ratio,
            attention_dropout_ratio=attention_dropout_ratio,
            drop_path_ratio=drop_path_ratio,
            norm_layer=norm_layer,
            act_layer=act_layer
        )
        self.norm = norm_layer(hidden_dim)
        self.head = nn.Linear(hidden_dim, num_classes)
        self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=.01)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.LayerNorm):
                nn.init.zeros_(m.bias)
                nn.init.ones_(m.weight)

    def forward(self, x):
        x = self.patch_embed(x)  # [B, 196, 768]
        cls_token = self.cls_token.expand(x.shape[0], -1, -1)  # [B, 1, 768]
        x = torch.cat((cls_token, x), dim=1)  # [B, 196+1, 768]
        x = self.pos_drop(x + self.pos_embed)  # [B, 197, 768]
        x = self.blocks(x)  # [B, 197, 768]
        x = x[:, 0]  # [B, 768]
        x = self.head(x)  # [B, num_classes]
        return x


def vit_b_16(num_classes=1000) -> VisionTransformer:
    return VisionTransformer(
        img_size=224,
        patch_size=16,
        num_classes=num_classes,
        hidden_dim=768,
        num_heads=12,
        num_layers=12,
    )

def vit_b_32(num_classes=1000) -> VisionTransformer:
    return VisionTransformer(
        img_size=224,
        patch_size=32,
        num_classes=num_classes,
        hidden_dim=768,
        num_heads=12,
        num_layers=12,
    )

def vit_l_16(num_classes=1000) -> VisionTransformer:
    return VisionTransformer(
        img_size=224,
        patch_size=16,
        num_classes=num_classes,
        hidden_dim=1024,
        num_heads=16,
        num_layers=24,
    )

def vit_l_32(num_classes=1000) -> VisionTransformer:
    return VisionTransformer(
        img_size=224,
        patch_size=32,
        num_classes=num_classes,
        hidden_dim=1024,
        num_heads=16,
        num_layers=24,
    )

def vit_h_14(num_classes=1000) -> VisionTransformer:
    return VisionTransformer(
        img_size=224,
        patch_size=14,
        num_classes=num_classes,
        hidden_dim=1280,
        num_heads=16,
        num_layers=32,
    )


if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = vit_b_16(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # vit_b_16 Total params: 85,651,204
    # vit_b_32 Total params: 87,420,676
    # vit_l_16 Total params: 303,105,028
    # vit_l_32 Total params: 305,464,324
    # vit_h_14 Total params: 630,442,244


"""
这种重参数化的重聚焦卷积（RefConv）方法是在已经有一个预训练的卷积神经网络（CNN）模型的基础上进行的。
具体来说，RefConv通过将可训练的重新聚焦变换应用于预训练模型的卷积核，以建立卷积核之间的连接，从而增强了模型的先验
Paper address: https://arxiv.org/pdf/2310.10563.pdf
The author provides the implementation of the pytorch algorithm at the end of the paper
"""
import torch
import torch.nn as nn
from torch.nn import functional as F

class RefConv(nn.Module):
    """
    Implementation of RefConv.
        --in_channels: number of input channels in the basis kernel
        --out_channels: number of output channels in the basis kernel
        --kernel_size: size of the basis kernel
        --stride: stride of the original convolution
        --padding: padding added to all four sides of the basis kernel
        --groups: groups of the original convolution
        --map_k: size of the learnable kernel
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride,
                 padding=None, groups=1, map_k=3):
        super(RefConv, self).__init__()
        assert map_k <= kernel_size
        self.origin_kernel_shape = (out_channels, in_channels // groups, kernel_size, kernel_size)
        self.register_buffer('weight', torch.zeros(*self.origin_kernel_shape))
        G = in_channels * out_channels // (groups ** 2)
        self.num_2d_kernels = out_channels * in_channels // groups
        self.kernel_size = kernel_size
        self.convmap = nn.Conv2d(in_channels=self.num_2d_kernels,
                                 out_channels=self.num_2d_kernels, kernel_size=map_k, stride=1, padding=map_k // 2,
                                 groups=G, bias=False)

        self.bias = None
        self.stride = stride
        self.groups = groups
        if padding is None:
            padding = kernel_size // 2
        self.padding = padding

    def forward(self, inputs):
        origin_weight = self.weight.view(1, self.num_2d_kernels, self.kernel_size, self.kernel_size)
        kernel = self.weight + self.convmap(origin_weight).view(*self.origin_kernel_shape)
        return F.conv2d(inputs, kernel, stride=self.stride, padding=self.padding,
                        dilation=1, groups=self.groups, bias=self.bias)


if __name__ == '__main__':
    block1 = RefConv(64, 32, 3, 1)
    block2 = nn.Conv2d(64, 32, 3, 1, padding=1)
    input = torch.randn((3, 64, 64, 64))
    output = block1(input)
    print(output.shape)
    output = block2(input)
    print(output.shape)


"""
Copyright (c) 2024, Auorui.
All rights reserved.

主要针对于opencv格式的, 一般这部分用于传统图像的增强
"""
import cv2
import random
import numpy as np
from math import ceil

def crop_image_by_2points(image, StartPoint, EndPoint):
    """
    Crop the image based on the starting and ending points
    Args:
        image: The image to be cropped
        StartPoint: Top left corner coordinates (x, y) of the cropping area
        EndPoint: The coordinates of the bottom right corner of the cropping area (x1, y1)

    Returns:Cropped image
    """
    height, width = image.shape[:2]
    x_min, y_min = StartPoint
    x_max, y_max = EndPoint

    assert x_max > x_min, "Maximum value of cropping x_max cannot be less than the minimum value x_min"
    assert y_max > y_min, "Maximum value of cropping y_max cannot be less than the minimum value y_min"
    assert x_min >= 0, "x_min cannot be less than 0"
    assert x_max <= width, "x_max cannot be greater than the image width"
    assert y_min >= 0, "y_min cannot be less than 0"
    assert y_max <= height, "y_max cannot be greater than the image height"

    return image[y_min:y_max, x_min:x_max]


def crop_image_by_1points(image, StartPoint, width, height):
    """
    Crop the image based on the starting point and specified width and height
    Args:
        image: The image to be cropped
        StartPoint: Top left corner coordinates (x, y) of the cropping area
        width: Width of cropping area
        height: Height of cropping area

    Returns: Cropped image
    """
    x_start, y_start = StartPoint
    assert width > 0 and height > 0, "Width and height of cropping area must be greater than 0"
    assert x_start >= 0 and y_start >= 0, "x_min and y_min cannot be less than 0"

    return image[y_start:y_start+height, x_start:x_start+width]

def center_crop(image, target_size):
    """
    Center-crops an image to the specified target size.

    Args:
        image (numpy.ndarray): The input image.
        target_size (tuple): A tuple (width, height) specifying the target size.

    Returns:
        numpy.ndarray: The center-cropped image.
    """
    h, w = image.shape[:2]
    crop_width, crop_height = target_size

    if crop_width > w or crop_height > h:
        raise ValueError("center_crop: Target size is larger than the input image size")

    x_start = (w - crop_width) // 2
    y_start = (h - crop_height) // 2

    return crop_image_by_1points(image, (x_start, y_start), crop_width, crop_height)

def five_crop(image, size):
    """
    Generate 5 cropped images (one central and four corners).

    Args:
        image (numpy.ndarray): Input image as a NumPy array.
        size (tuple): A tuple (width, height) specifying the crop size.

    Returns:
        list: A list of 5 NumPy arrays.
    """
    width, height = image.shape[1], image.shape[0]
    crop_width, crop_height = size

    if crop_width > width or crop_height > height:
        raise ValueError("Crop size exceeds the image dimensions.")

    center_x = width // 2
    center_y = height // 2

    crops = []

    # Central crop
    left = center_x - crop_width // 2
    upper = center_y - crop_height // 2
    right = left + crop_width
    lower = upper + crop_height
    central_crop = image[upper:lower, left:right]
    crops.append(central_crop)

    # Top-left corner crop
    top_left_crop = image[0:crop_height, 0:crop_width]
    crops.append(top_left_crop)

    # Top-right corner crop
    top_right_crop = image[0:crop_height, width - crop_width:width]
    crops.append(top_right_crop)

    # Bottom-left corner crop
    bottom_left_crop = image[height - crop_height:height, 0:crop_width]
    crops.append(bottom_left_crop)

    # Bottom-right corner crop
    bottom_right_crop = image[height - crop_height:height, width - crop_width:width]
    crops.append(bottom_right_crop)

    return crops

def stitcher_image(image_paths: list):
    """
    Image stitching, smaller images may result in stitching failure
    Args:
        image_paths: A list composed of image paths

    Returns: the stitched image
    """
    opencv_version = cv2.__version__
    major_version = int(opencv_version.split(".")[0])
    if major_version >= 4:
        stitcher = cv2.Stitcher.create() # 适用于OpenCV 4.x
    else:
        stitcher = cv2.createStitcher()  # 适用于OpenCV 3.x 或更早版本
    images = []
    for path in image_paths:
        img = cv2.imread(path)
        if img is not None:
            images.append(img)
    assert len(images) >= 2, "At least two images are required for stitching"
    (status, stitched_image) = stitcher.stitch(images)
    assert status == cv2.Stitcher_OK, 'Image stitching failed'
    return stitched_image

def centerzoom(image, zoom_factor: float):
    """
    Center Zoom Image
    Args:
        image: The image to be scaled
        zoom_factor: Scale factor

    Returns: the image scaled by the center

    """
    h, w = image.shape[:2]
    h_ch, w_ch = ceil(h / zoom_factor), ceil(w / zoom_factor)
    h_top, w_top = (h - h_ch) // 2, (w - w_ch) // 2
    zoomed_img = cv2.resize(image[h_top : h_top + h_ch, w_top : w_top + w_ch], (w, h), interpolation=cv2.INTER_LINEAR)

    return zoomed_img


def flip(image, option_value):
    """
    Flip the image
    Args:
        image: numpy array of image
        option_value: random integer between 0 to 2
            vertical                          0
            horizontal                        1
            horizontally and vertically flip  2
    Returns: numpy array of flipped image
    """
    if option_value == 0:
        image = np.flip(image, option_value)
    elif option_value == 1:
        image = np.flip(image, option_value)
    elif option_value == 2:
        image = np.flip(image, 0)
        image = np.flip(image, 1)
    else:
        image = image

    return image

def horizontal_flip(image):
    """Flip the image horizontally"""
    return flip(image, 1)

def vertical_flip(image):
    """Flip the image vertically"""
    return flip(image, 0)

def resize(image, size=None, scale=None):
    """
    bilinear interpolation: https://blog.csdn.net/m0_62919535/article/details/132094815
    Args:
        image (numpy.ndarray): Image to be resized.
        size (tuple): New size in the format (width, height).
        scale (float): Scale of image
    Returns:
        Resized image.
    """
    if size is None and scale is None:
        raise ValueError("Either size or scale must be provided.")

    if scale is not None:
        ah, aw, channel = image.shape
        bh, bw = int(ah * scale), int(aw * scale)
        dst_img = np.zeros((bh, bw, channel), np.uint8)

        y_coords, x_coords = np.meshgrid(np.arange(bh), np.arange(bw), indexing='ij')
        AX = (x_coords + 0.5) / scale - 0.5
        AY = (y_coords + 0.5) / scale - 0.5

        x1 = np.floor(AX).astype(int)
        y1 = np.floor(AY).astype(int)
        x2 = np.minimum(x1 + 1, aw - 1)
        y2 = np.minimum(y1 + 1, ah - 1)
        R1 = ((x2 - AX)[:, :, np.newaxis] * image[y1, x1]).astype(float) + (
                (AX - x1)[:, :, np.newaxis] * image[y1, x2]).astype(float)
        R2 = ((x2 - AX)[:, :, np.newaxis] * image[y2, x1]).astype(float) + (
                (AX - x1)[:, :, np.newaxis] * image[y2, x2]).astype(float)

        dst_img = (y2 - AY)[:, :, np.newaxis] * R1 + (AY - y1)[:, :, np.newaxis] * R2

    if size is not None:
        dst_img = cv2.resize(image, size, interpolation=cv2.INTER_LINEAR)

    return dst_img.astype(np.uint8)

def adjust_brightness_cv2(image, brightness_factor):
    """
    Adjust brightness of an image using OpenCV.
    Args:
        image (numpy.ndarray): Image to be adjusted.
        brightness_factor (float): A factor by which to adjust brightness.
            - 0.0 gives a black image.
            - 1.0 gives the original image.
            - Greater than 1.0 increases brightness.
            - Less than 1.0 decreases brightness.
    Returns:
        Brightness-adjusted image.
    """
    return cv2.convertScaleAbs(image, alpha=brightness_factor, beta=0)

def adjust_brightness_numpy(image, brightness_factor):
    """
    Adjust brightness of an image using Numpy.
    Args:
        image (numpy.ndarray): Image to be adjusted.
        brightness_factor (float): A factor by which to adjust brightness.
            - 0.0 gives a black image.
            - 1.0 gives the original image.
            - Greater than 1.0 increases brightness.
            - Less than 1.0 decreases brightness.
    Returns:
        Brightness-adjusted image.
    """
    image_float = image.astype(np.float32)
    _image = image_float * brightness_factor
    _image = np.clip(_image, 0, 255)
    b_image = _image.astype(np.uint8)
    return b_image

def rotate(image, angle, fill_value=(0, 0, 0)):
    """
    Rotate the input image by angle degrees.

    Args:
        image (numpy.ndarray): Input image as a NumPy array.
        angle (float): Rotation angle in degrees, counter-clockwise.
        fill_value (tuple, optional): Fill color for areas outside the rotated image.
            Default is (0, 0, 0) for black.

    Returns:
        numpy.ndarray: Rotated image.

    """
    height, width = image.shape[:2]
    rotation_matrix = cv2.getRotationMatrix2D((width / 2, height / 2), angle, 1)
    rotated_image = cv2.warpAffine(image, rotation_matrix, (width, height), borderValue=fill_value)
    return rotated_image

def adjust_gamma(image, gamma, gain=1):
    """
    Adjust the gamma correction and gain of an input OpenCV image.

    Gamma correction is a nonlinear operation used to encode or decode luminance
    or tristimulus values in video or still image systems. Increasing gamma
    values make the dark parts of an image darker and the bright parts brighter,
    while decreasing gamma values have the opposite effect. The gain parameter
    scales the intensity of the image after gamma correction.

    Args:
        image (numpy.ndarray): Input image as a NumPy array, typically in the range [0, 255]
            for 8-bit images.
        gamma (float): Gamma correction factor. Values typically range from 0.5 to 2.5.
            A value of 1.0 leaves the image unchanged.
        gain (float, optional): Gain factor to scale the image after gamma correction.
            Defaults to 1.0 (no change in intensity).

    Returns:
        numpy.ndarray: Gamma-corrected and gain-adjusted image as a NumPy array, with
            pixel values in the range [0, 255].
    """
    img_np = np.array(image)
    img_gamma_corrected = ((img_np / 255.0) ** gamma) * 255.0 * gain
    img_gamma_corrected = np.clip(img_gamma_corrected, 0, 255).astype(np.uint8)
    return img_gamma_corrected

def pad_margin(image, padding, fill_value=(128, 128, 128)):
    """
    Add margins to the input image.
    Args:
        image: Input the image as a NumPy array.
        padding: Margin size.
            -If it is an integer, all margins (top, bottom, left, right) are set to that value.
            -If it is a tuple or list of length 2, the first element represents the left and right margins, and the second element represents the top and bottom margins.
            -If it is a tuple or list of length 4, it represents the left, top, right, and bottom margins in order.
        fill_value: Fill color values (BGR format)
    Returns: Image with added margins
    """
    if not isinstance(padding, (int, tuple, list)):
        raise TypeError("Padding should be an integer or a tuple/list of two or four values.")

    if not isinstance(fill_value, (int, str, tuple)):
        raise TypeError("Fill value should be an integer, a string, or a tuple.")

    top = bottom = left = right = None
    if isinstance(padding, int):
        top = bottom = left = right = padding
    elif len(padding) == 2:
        left = right = padding[0]
        top = bottom = padding[1]
    elif len(padding) == 4:
        left = padding[0]
        top = padding[1]
        right = padding[2]
        bottom = padding[3]

    img_np = np.array(image)
    if img_np.shape[-1] == 1:
        fill_value = fill_value[0]
        # BORDER_CONSTANT  BORDER_ISOLATED
    img_np = cv2.copyMakeBorder(img_np, top, bottom, left, right, cv2.BORDER_CONSTANT, value=fill_value)

    return img_np

def erase(image, StartPoint, height, width, erase_value=(128,128,128)):
    """
    Erase a rectangular region in a NumPy image array.

    Args:
        image (numpy.ndarray): Input NumPy image array.
        StartPoint (tuple): Top left corner coordinates (x, y) of the cropping area
        x (int): X-coordinate of the top-left corner of the region to be erased.
        y (int): Y-coordinate of the top-left corner of the region to be erased.
        height (int): Height of the erased region.
        width (int): Width of the erased region.
        erase_value (tuple, optional): The RGB color value to fill the erased region with.
            Default is (128, 128, 128), which corresponds to gray.

    Returns:
        numpy.ndarray: NumPy image array with the specified region erased and filled with the erase_value.
    """
    if not isinstance(image, np.ndarray):
        raise TypeError("np_img should be a NumPy array. Got {}.".format(type(image)))
    x, y = StartPoint
    erased_img = image.copy()
    erased_img[y:y + height, x:x + width, :] = erase_value
    return erased_img

def enhance_hsv(image, hgain=0.5, sgain=0.5, vgain=0.5):
    """
    HSV color space enhancement
    Args:
        image (numpy.ndarray):  The input image should be a NumPy array in BGR format.
        Hgain (float): The gain value of Hue (color tone). The default value is 0.5.
        Sgain (float): The gain value of saturation. The default value is 0.5.
        Vgain (float): The gain value of Value (brightness). The default value is 0.5.
            -The gain value can be any real number, but is usually set between -1 and 1.
            -Positive values will enhance the corresponding color components, while negative values will weaken.

    Returns: Enhanced HSV color space image, returned in BGR format
    """
    if hgain or sgain or vgain:
        r = np.random.uniform(-1, 1, 3) * [hgain, sgain, vgain] + 1  # random gains
        hue, sat, val = cv2.split(cv2.cvtColor(image, cv2.COLOR_BGR2HSV))
        dtype = image.dtype  # uint8

        x = np.arange(0, 256, dtype=r.dtype)
        lut_hue = ((x * r[0]) % 180).astype(dtype)
        lut_sat = np.clip(x * r[1], 0, 255).astype(dtype)
        lut_val = np.clip(x * r[2], 0, 255).astype(dtype)

        im_hsv = cv2.merge((cv2.LUT(hue, lut_hue), cv2.LUT(sat, lut_sat), cv2.LUT(val, lut_val)))
        return cv2.cvtColor(im_hsv, cv2.COLOR_HSV2BGR, dst=image)

def hist_equalize(image, clahe=True, is_bgr=True):
    """
    Perform histogram equalization on the brightness channels of the image to enhance its contrast.
    You can choose to use regular histogram equalization (cv2. equalizeHist) or contrast limited adaptive histogram equalization (CLAHE, cv2. create CLAHE).

    Args:
        image (numpy.ndarray):  The input image should be a NumPy array in BGR or RGB format.
        clahe (bool):  Should we use Contrast Constrained Adaptive Histogram Equalization (CLAHE). The default is True.
                    -If True, use CLAHE for equalization.
                    -If False, use regular histogram equalization.
        is_bgr (bool):  Is the input image in BGR format. The default is True.
                    -If True, assume the input image is in BGR format and convert the color space accordingly.
                    -If False, assume the input image is in RGB format and convert the color space accordingly.

    Returns:
        numpy.ndarray: The image obtained by histogram equalization of the brightness
        channel is returned in BGR or RGB format, depending on the format of the input image.
    """
    # 均衡BGR图像“im”上的直方图，其形状为im(n，m，3)，范围为0-255
    yuv = cv2.cvtColor(image, cv2.COLOR_BGR2YUV if is_bgr else cv2.COLOR_RGB2YUV)
    if clahe:
        c = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
        yuv[:, :, 0] = c.apply(yuv[:, :, 0])
    else:
        yuv[:, :, 0] = cv2.equalizeHist(yuv[:, :, 0])
    return cv2.cvtColor(yuv, cv2.COLOR_YUV2BGR if is_bgr else cv2.COLOR_YUV2RGB)

def random_lighting(image, alpha):
    """
    Add AlexNet-style PCA-based noise to an image.
    Args:
        image (numpy.ndarray):  The input image should be a NumPy array in BGR format.
        alpha:(float):  Control the standard deviation of noise amplitude.
                        The larger the alpha value, the greater the added noise.

    Returns:
        numpy.ndarray:  The image with PCA noise added is returned in BGR format with data type uint8.
    """
    alpha_b = np.random.normal(loc=0.0, scale=alpha)
    alpha_g = np.random.normal(loc=0.0, scale=alpha)
    alpha_r = np.random.normal(loc=0.0, scale=alpha)
    table = np.array([
        [55.46 * -0.5675, 4.794 * 0.7192, 1.148 * 0.4009],
        [55.46 * -0.5808, 4.794 * -0.0045, 1.148 * -0.8140],
        [55.46 * -0.5836, 4.794 * -0.6948, 1.148 * 0.4203]
    ])
    pca_b = table[2][0] * alpha_r + table[2][1] * alpha_g + table[2][2] * alpha_b
    pca_g = table[1][0] * alpha_r + table[1][1] * alpha_g + table[1][2] * alpha_b
    pca_r = table[0][0] * alpha_r + table[0][1] * alpha_g + table[0][2] * alpha_b
    img_arr = np.array(image).astype(np.float64)
    img_arr[:, :, 0] += pca_b
    img_arr[:, :, 1] += pca_g
    img_arr[:, :, 2] += pca_r
    img_arr = np.uint8(np.minimum(np.maximum(img_arr, 0), 255))

    return img_arr

def random_rotation(image, degrees=None, fill_value=(0, 0, 0)):
    """
    Randomly rotate the image based on whether it is a square and the provided degree range.
    Args:
        image (numpy.ndarray):  The input image should be a NumPy array in BGR format.
        degrees (tuple, optional):  The range of rotation angles is in the format of (min_degree, max_degree). If None, select the default angle based on the shape of the image. The default is None.
        fill_value (tuple, optional):  The color filled in the blank area at the edge of the rotated image is in the format of (B, G, R). The default is (0,0,0), which is black.

    Returns:
        numpy.ndarray:  The rotated image is returned in BGR format.
    """
    h, w = image.shape[:2]
    is_squre = True if h == w else False
    if degrees is not None:
        angle = random.uniform(degrees[0], degrees[1])
    else:
        if is_squre:
            angle = random.choice([0, 90, 180, 270])
        else:
            angle = random.choice([0, 180])
    return rotate(image, angle, fill_value)

def random_rot90(image):
    """Randomly rotate the input image by 90 degrees"""
    r = random.randint(0, 3)
    return np.rot90(image, r, (0, 1))

def random_horizontal_flip(image, prob=0.5):
    """Randomly flip the input image horizontally."""
    if prob > random.random():
        image = horizontal_flip(image)
    return image

def random_vertical_flip(image, prob=0.5):
    """Randomly flip the input image vertically."""
    if prob > random.random():
        image = vertical_flip(image)
    return image

def random_crop(image, crop_size):
    """
    Randomly crop an image to the specified size.

    Args:
        image (numpy.ndarray): Input image as a NumPy array.
        crop_size (tuple): A tuple (width, height) specifying the crop size.

    Returns:
        Randomly cropped image.
    """
    height, width = image.shape[:2]
    max_x = width - crop_size[0]
    max_y = height - crop_size[1]

    start_x = random.randint(0, max_x)
    start_y = random.randint(0, max_y)

    cropped_image = crop_image_by_1points(image, (start_x, start_y), crop_size[0], crop_size[1])

    return cropped_image


def random_resize_crop(image, target_size, scale_range=(1., 2.)):
    """
    Randomly resize and crop an image.

    Args:
        image (numpy.ndarray): Input image as a NumPy array.
        target_size (tuple): A tuple (width, height) specifying the target size.
        scale_range (tuple, optional): A tuple (min_scale, max_scale) specifying the range of scaling.
            Default is (0.8, 1.2), which allows resizing between 80% and 120% of the original size.

    Returns:
        Randomly resized and cropped image.
    """
    min_scale, max_scale = scale_range
    scale_factor = random.uniform(min_scale, max_scale)

    height, width = image.shape[:2]

    new_width = int(width * scale_factor)
    new_height = int(height * scale_factor)

    resized_image = resize(image, (new_width, new_height))

    crop_x = random.randint(0, new_width - target_size[0])
    crop_y = random.randint(0, new_height - target_size[1])
    cropped_image = crop_image_by_1points(resized_image, (crop_x, crop_y), target_size[0], target_size[1])

    return cropped_image


def resizepad(image, target_shape, frame=True):
    """
    Adjust the image size and perform grayscale filling
    Args:
        image (numpy.ndarray): Input image as a NumPy array.
        target_shape (tuple): A tuple (height, width) specifying the target shape.
        frame (bool): Do you want to resize without distortion

    Returns:
        the adjusted image
    """
    h, w = target_shape
    ih, iw = image.shape[:2]
    if frame:
        scale = min(w / iw, h / ih)
        nw = int(iw * scale)
        nh = int(ih * scale)
        resized_image = cv2.resize(image, (nw, nh), interpolation=cv2.INTER_CUBIC)
        new_image = np.full((h, w, 3), (128, 128, 128), dtype=np.uint8)
        top = (h - nh) // 2
        left = (w - nw) // 2
        new_image[top:top + nh, left:left + nw] = resized_image
        return new_image, nw, nh
    else:
        new_image = cv2.resize(image, (w, h), interpolation=cv2.INTER_CUBIC)
        return new_image

def addnoisy(image, n=10000):
    """
    Add salt and pepper noise to the image.

    Args:
        image (numpy.ndarray):  The input image should be a NumPy array of grayscale or color images (BGR or RGB format).
        n (int, optional):  The number of salt and pepper noise points to be added. The default is 10000.

    Returns: The image after adding salt and pepper noise has the same format as the input image.
    """
    result = image.copy()
    h, w = image.shape[:2]
    for i in range(n):
        x = np.random.randint(0, w)
        y = np.random.randint(0, h)
        if np.random.randint(0, 2) == 0:
            result[x, y] = 0
        else:
            result[x, y] = 255
    return result

def addfog(image, beta=0.05, brightness=0.5, use_efficient=True):
    """
    Efficient implementation of adding haze effect to input images based on atmospheric scattering model.

    Args:
        image (numpy.ndarray): The input image ranges from 0-255
        beta (float, optional): Parameters for controlling haze effects The larger the beta
            value, the more pronounced the haze effect The default value is 0.05
        brightness (float, optional): The brightness value of haze The larger the value,
            the higher the overall brightness of the image The default is 0.5
        use_efficient (bool, optional): Whether to adopt a more efficient approach, default to True
    Returns:
        numpy.ndarray: Image with haze effect added, data type uint8, range 0-255。
    """
    img_f = image.astype(np.float32) / 255.0
    row, col, chs = image.shape
    size = np.sqrt(max(row, col))  # Atomization size
    center = (row // 2, col // 2)  # Atomization center
    if use_efficient:
        y, x = np.ogrid[:row, :col]
        dist = np.sqrt((x - center[1])**2 + (y - center[0])**2)
        d = -0.04 * dist + size
        td = np.exp(-beta * d)
        img_f = img_f * td[..., np.newaxis] + brightness * (1 - td[..., np.newaxis])
    else:
        for j in range(row):
            for l in range(col):
                d = -0.04 * np.sqrt((j - center[0]) ** 2 + (l - center[1]) ** 2) + size
                td = np.exp(-beta * d)
                img_f[j][l][:] = img_f[j][l][:] * td + brightness * (1 - td)
    img_f = np.clip(img_f * 255, 0, 255).astype(np.uint8)
    return img_f

def addfog_channels(image, fog_intensity=0.5, fog_color_intensity=255):
    """
    对图像 RGB 通道应用雾效。

    参数:
        image: 输入图像（numpy数组）。
        fog_intensity: 雾的强度（0到1）。
        fog_color_intensity: 雾的颜色强度（0到255）.不宜过小, 建议大于180
    """
    fog_intensity = np.clip(fog_intensity, 0, 1)
    fog_layer = np.ones_like(image) * fog_color_intensity
    fogged_image = cv2.addWeighted(image, 1 - fog_intensity, fog_layer, fog_intensity, 0)

    return fogged_image

class Retinex():
    """
    增强算法Retinex
    - 单尺度 : SSR
    - 多尺度 : MSR
    - 多尺度自适应增益 : MSRCR
    https://blog.csdn.net/m0_62919535/article/details/130372571
    """
    def SSR(self, img, sigma):
        """
        将输入图像转换为了对数空间。/255将像素值归一化到0到1之间，np.log1p取对数并加1是为了避免出现对数运算中分母为0的情况。二维离散傅里叶变换将
        图像从空间域变换到频率域，可以提取出图像中的频率信息。G_recs是用于计算高斯核的半径，result用于最后三通道的叠加。然后循环用于计算加权后的
        频率域图像，再逆二维离散傅里叶变换，得到反射图像，对反射图像进行指数变换，得到最终的输出图像。
        :param img: 输入图像
        :param sigma: 高斯分布的标准差
        :return:
        """
        img_log = np.log1p(np.array(img, dtype="float") / 255)
        img_fft = np.fft.fft2(img_log)
        G_recs = sigma // 2 + 1
        result = np.zeros_like(img_fft)
        rows, cols, deep = img_fft.shape
        for z in range(deep):
            for i in range(rows):
                for j in range(cols):
                    for k in range(1, G_recs):
                        G = np.exp(-((np.log(k) - np.log(sigma)) ** 2) / (2 * np.log(2) ** 2))
                        #计算高斯滤波器的权值，其中sigma是高斯分布的标准差，k是高斯滤波器的半径，G是高斯滤波器在该点的权值。
                        result[i, j] += G * img_fft[i, j]
        img_ssr = np.real(np.fft.ifft2(result))
        img_ssr = np.exp(img_ssr) - 1
        img_ssr = np.uint8(cv2.normalize(img_ssr, None, 0, 255, cv2.NORM_MINMAX))
        return img_ssr

    def MSR(self, img, scales):
        """
        MSR算法在图像增强中与SSR不同的是，它不需要进行频域变换，它主要是基于图像在多个尺度下的平滑处理和差分处理来提取图像的局部对比度信息和全
        局对比度信息，从而实现对图像的增强。
        在 MSR 算法中，先对图像进行对数变换得到对数图像，然后在不同的尺度下，使用高斯滤波对图像进行平滑处理，得到不同尺度下的平滑图像。接着，通
        过将对数图像和不同尺度下的平滑图像进行差分，得到多个尺度下的细节图像。最后，将这些细节图像加权融合，输出最终的增强图像。

        :param img:
        :param scales: 取值大概在1-10之间
        :return:
        """
        img_log = np.log1p(np.array(img, dtype="float") / 255)
        result = np.zeros_like(img_log)
        img_light = np.zeros_like(img_log)
        r, c, deep = img_log.shape
        for z in range(deep):
            for scale in scales:
                kernel_size = scale * 4 + 1
                # 高斯滤波器的大小，经验公式kernel_size = scale * 4 + cat
                sigma = scale
                img_smooth = cv2.GaussianBlur(img_log[:, :, z], (kernel_size, kernel_size), sigma)
                img_detail = img_log[:, :, z] - img_smooth
                result[:, :, z] += cv2.resize(img_detail, (c, r))
                img_light[:, :, z] += cv2.resize(img_smooth, (c, r))
        img_msr = np.exp(result+img_light) - 1
        img_msr = np.uint8(cv2.normalize(img_msr, None, 0, 255, cv2.NORM_MINMAX))
        return img_msr


    def MSRCR(self, img, scales, k):
        """
        使用 MSRCR 算法的步骤：
            1. 将输入图像转换为对数空间（log-space）。
            2. 对图像在不同尺度下进行平滑处理，得到不同尺度下的平滑图像。
            3. 计算细节图像，通过将对数图像与平滑图像相减获得。
            4. 对细节图像进行缩放，以匹配输入图像的大小。
            5. 根据尺度因子调整细节图像的权重。
            6. 将权重调整后的细节图像与平滑图像相加，得到增强的图像。
            7. 最后，将增强的图像反转对数变换，以还原原始图像的像素值范围。
        :param img:
        :param scales:取值大概在1-10之间
        :param k: k的取值范围在10~20之间比较合适。当k取值较小时，图像的细节增强效果比较明显，但会出现较强的噪点，当k取值较大时，图像的细节
                    增强效果不明显，但噪点会减少。
        :return:
        """
        img_log = np.log1p(np.array(img, dtype="float") / 255)
        result = np.zeros_like(img_log)
        img_light = np.zeros_like(img_log)
        r, c, deep = img_log.shape
        for z in range(deep):
            for scale in scales:
                kernel_size = scale * 4 + 1
                # 高斯滤波器的大小，经验公式kernel_size = scale * 4 + cat
                sigma = scale
                G_ratio=sigma**2/(sigma**2+k)
                img_smooth = cv2.GaussianBlur(img_log[:, :, z], (kernel_size, kernel_size), sigma)
                img_detail = img_log[:, :, z] - img_smooth
                result[:, :, z] += cv2.resize(img_detail, (c, r))
                result[:, :, z]=result[:, :, z]*G_ratio
                img_light[:, :, z] += cv2.resize(img_smooth, (c, r))

        img_msrcr = np.exp(result+img_light) - 1
        img_msrcr = np.uint8(cv2.normalize(img_msrcr, None, 0, 255, cv2.NORM_MINMAX))
        return img_msrcr

if __name__=="__main__":
    image_path = r"D:\PythonProject\pyzjrPyPi\pyzjr\data\outdoor.png"
    image = cv2.imread(image_path)
    print(image.shape)
    # process_img = crop_image_by_2points(image, StartPoint=(200, 20), EndPoint=(400, 100))
    # process_img = crop_image_by_1points(image, StartPoint=(200, 20), width=100, height=200)
    # process_img = center_crop(image, target_size=(200, 200))
    # process_img = five_crop(image, (100, 100))[0]
    # process_img = centerzoom(image, 1.2)
    # process_img = flip(image, 0)
    # process_img = vertical_flip(image)
    # process_img = horizontal_flip(image)
    # process_img = resize(image, (200, 200), scale=1.2)
    # process_img = adjust_brightness_cv2(image, 0.5)
    # process_img = adjust_brightness_numpy(image, 0.5)
    # process_img = rotate(image, 20)
    # process_img = adjust_gamma(image, 1.2, 1.2)
    # process_img = pad_margin(image, padding=(24, 25))
    # process_img = erase(image, (100, 100), 120, 130)
    # process_img = enhance_hsv(image)
    # process_img = hist_equalize(image, clahe=False)
    # process_img = random_lighting(image, alpha=0.7)
    # process_img = random_rotation(image)
    process_img = random_rot90(image)
    # process_img = random_horizontal_flip(image)
    # process_img = random_vertical_flip(image)
    # process_img = random_crop(image, (250, 250))
    # process_img = random_resize_crop(image, (250, 250))
    # process_img, _, _ = resizepad(image, (256, 256))
    # process_img = addnoisy(image, n=10000)
    # process_img = addfog(image)
    # process_img = addfog_channels(image)
    print(process_img.shape)
    cv2.imshow("ss", process_img)
    cv2.waitKey(0)



# 像素级别的研究

from .pixel import (
    SkeletonMap,
    incircle,
    outcircle,
    foreground_contour_length,
    get_each_crack_areas
)

from .steger import (
    Steger,
    Magnitudefield,
    derivation,
    nonMaxSuppression,
)

from .skeleton_extraction import (
    skeletoncv,
    sketionio,
    sketion_medial_axis
)

from .crack import *

from .quality_eval import (
    calculate_ssim,
    calculate_ssimv2,
    calculate_psnr,
    calculate_psnrv2,
    gradient_metric,
    laplacian_metric
)




__version__ = '1.4.4'

import os
import torch
import sys
import platform

def pyzjr_info():
    print("Python版本:", ".".join(map(str, [sys.version_info.major,
                                          sys.version_info.minor,
                                          sys.version_info.micro])))
    print("PyTorch版本:", torch.__version__)
    if torch.cuda.is_available():
        print("CUDA版本:", torch.version.cuda)
        print("CUDA设备名称:", torch.cuda.get_device_name(0))
    print("CPU型号:", platform.processor())
    print("CPU核心数量:", os.cpu_count())
    print("系统信息:", platform.system() + " " + platform.version())
    print("---------------------------")
    print("Pyzjr版本:", __version__)

if __name__=="__main__":
    pyzjr_info()


import torch
import math
import warnings
from torch import nn
import torch.nn.init as init

def init_weights_complex(model, conv=None, batchnorm=None, linear=None, lstm=None):
    """
    :param model: Pytorch Model which is nn.Module
    :param conv:  'kaiming' or 'xavier'
    :param batchnorm: 'normal' or 'constant'
    :param linear: 'kaiming' or 'xavier'
    :param lstm: 'kaiming' or 'xavier'
    """
    for m in model.modules():
        if conv is not None and isinstance(m, (nn.modules.conv._ConvNd)):
            if conv == 'kaiming':
                init.kaiming_normal_(m.weight)
            elif conv == 'xavier':
                init.xavier_normal_(m.weight)
            else:
                raise ValueError("init type of conv error.\n")
            if m.bias is not None:
                init.constant_(m.bias, 0)

        elif batchnorm is not None and isinstance(m, (nn.modules.batchnorm._BatchNorm)):
            if batchnorm == 'normal':
                init.normal_(m.weight, 1.0, 0.02)
            elif batchnorm == 'constant':
                init.constant_(m.weight, 1.0)
            else:
                raise ValueError("init type of batchnorm error.\n")
            init.constant_(m.bias, 0.0)

        elif linear is not None and isinstance(m, nn.Linear):
            if linear == 'kaiming':
                init.kaiming_normal_(m.weight)
            elif linear == 'xavier':
                init.xavier_normal_(m.weight)
            else:
                raise ValueError("init type of linear error.\n")
            if m.bias is not None:
                init.constant_(m.bias, 0)

        elif lstm is not None and isinstance(m, nn.LSTM):
            for name, param in m.named_parameters():
                if 'weight' in name:
                    if lstm == 'kaiming':
                        init.kaiming_normal_(param)
                    elif lstm == 'xavier':
                        init.xavier_normal_(param)
                    else:
                        raise ValueError("init type of lstm error.\n")
                elif 'bias' in name:
                    init.constant_(param, 0)

def init_weights_simply(self):
    for m in self.modules():
        if isinstance(m, nn.Conv2d):
            init.kaiming_normal_(m.weight, mode='fan_out')
            if m.bias is not None:
                init.constant_(m.bias, 0)
        elif isinstance(m, nn.BatchNorm2d):
            init.constant_(m.weight, 1)
            init.constant_(m.bias, 0)
        elif isinstance(m, nn.Linear):
            init.normal_(m.weight, std=0.001)
            if m.bias is not None:
                init.constant_(m.bias, 0)

def official_init(model):
    # Official init from torch repo.
    for m in model.modules():
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal_(m.weight)
        elif isinstance(m, nn.BatchNorm2d):
            nn.init.constant_(m.weight, 1)
            nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.Linear):
            nn.init.constant_(m.bias, 0)

def xavier_init(module, gain=1, bias=0, distribution='normal'):
    assert distribution in ['uniform', 'normal']
    if hasattr(module, 'weight') and module.weight is not None:
        if distribution == 'uniform':
            nn.init.xavier_uniform_(module.weight, gain=gain)
        else:
            nn.init.xavier_normal_(module.weight, gain=gain)
    if hasattr(module, 'bias') and module.bias is not None:
        nn.init.constant_(module.bias, bias)

def normal_init(module, mean=0, std=1, bias=0):
    if hasattr(module, 'weight') and module.weight is not None:
        nn.init.normal_(module.weight, mean, std)
    if hasattr(module, 'bias') and module.bias is not None:
        nn.init.constant_(module.bias, bias)

def uniform_init(module, a=0, b=1, bias=0):
    if hasattr(module, 'weight') and module.weight is not None:
        nn.init.uniform_(module.weight, a, b)
    if hasattr(module, 'bias') and module.bias is not None:
        nn.init.constant_(module.bias, bias)

def trunc_normal_init(module: nn.Module,
                      mean: float = 0,
                      std: float = 1,
                      a: float = -2,
                      b: float = 2,
                      bias: float = 0) -> None:
    if hasattr(module, 'weight') and module.weight is not None:
        trunc_normal_(module.weight, mean, std, a, b)  # type: ignore
    if hasattr(module, 'bias') and module.bias is not None:
        nn.init.constant_(module.bias, bias)  # type: ignore

def _trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1. + math.erf(x / math.sqrt(2.))) / 2.

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn("mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
                      "The distribution of values may be incorrect.",
                      stacklevel=2)

    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the normal distribution.
    # Get upper and lower cdf values
    l = norm_cdf((a - mean) / std)
    u = norm_cdf((b - mean) / std)

    # Uniformly fill tensor with values from [l, u], then translate to
    # [2l-1, 2u-1].
    tensor.uniform_(2 * l - 1, 2 * u - 1)

    # Use inverse cdf transform for normal distribution to get truncated
    # standard normal
    tensor.erfinv_()

    # Transform to proper mean, std
    tensor.mul_(std * math.sqrt(2.))
    tensor.add_(mean)

    # Clamp to ensure it's in the proper range
    tensor.clamp_(min=a, max=b)
    return tensor


def trunc_normal_(tensor, mean=0., std=1., a=-2., b=2.):
    r"""Fills the input Tensor with values drawn from a truncated
    normal distribution. The values are effectively drawn from the
    normal distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`
    with values outside :math:`[a, b]` redrawn until they are within
    the bounds. The method used for generating the random values works
    best when :math:`a \leq \text{mean} \leq b`.

    NOTE: this impl is similar to the PyTorch trunc_normal_, the bounds [a, b] are
    applied while sampling the normal with mean/std applied, therefore a, b args
    should be adjusted to match the range of mean, std args.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff value
        b: the maximum cutoff value
    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.trunc_normal_(w)
    """
    with torch.no_grad():
        return _trunc_normal_(tensor, mean, std, a, b)


from .base import (
    natsorted,
    natural_key,
    timestr,
    alphabetstr,
    list_dirs,
    list_files,
    list_names
)
from .loaders import RepeatDataLoader, Repeat_sampler, seed_worker, \
    TrainDataloader, EvalDataloader

from .datasets import *

from .scripts import *
from .utils import *


from .losses import *
from .metrics import *
from .models import *
from .torchutils import *
from .strategy import *
from .save_pth import (
    SaveModelPth,
    SaveModelPthSimplify,
    SaveModelPthBestloss,
    SaveModelPthBestMetrics,
    load_partial_weights
)
from .tools import summary_1, summary_2, profile, time_sync
from .callbacks import (
    AverageMeter,
    SmoothedValue,
    MetricLogger,
    ConfusionMatrix,
    LossHistory,
    ErrorRateMonitor,
    ProcessMonitor,
)
from .devices import (
    Central_Processing_Unit,
    Graphics_Processing_Unit,
    use_all_gpu,
    num_worker,
    load_owned_device,
    load_device_on_model,
    CPU, cpu,
    GPU, gpu,
    release_gpu_memory,
    release_memory,
    PeakCPUMemory,
    start_measure,
    end_measure,
    log_measures
)



import numpy as np
import torch
import random
import warnings

# def mixup(img1, labels, img2, labels2):
#     # Applies MixUp augmentation https://arxiv.org/pdf/1710.09412.pdf
#     r = np.random.beta(32.0, 32.0)  # mixup ratio, alpha=beta=32.0
#     img = (img1 * r + img2 * (1 - r)).astype(np.uint8)
#     labels = np.concatenate((labels, labels2), 0)
#     return img, labels

def preprocess_input(image, mean=None, std=None):
    """
    将图像预处理为适合神经网络输入的格式。
    如果图像的最大值大于1，则先归一化到 [0, 1] 范围，
    然后根据提供的均值（mean）和标准差（std）进行标准化处理。

    :param image: 输入的图像，可以是 PIL 图像或 NumPy 数组格式。
    :param mean: 可选，图像标准化时使用的均值，默认为 None。
    :param std: 可选，图像标准化时使用的标准差，默认为 None。
    :return: 归一化和标准化后的图像。
    """
    image = np.asarray(image)
    if np.max(image) > 1:
        image = image / 255.0

    if mean is not None:
        mean = np.array(mean)
        image = image - mean

    if std is not None:
        std = np.array(std)
        image = image / std

    return image.astype(np.float32)

def SeedEvery(seed=11, rank=0, use_deterministic_algorithms=None):
    """
    :param seed: 设置基础随机种子
    :param rank: 进程的排名，用于为每个进程设置不同的种子
    :param use_deterministic_algorithms: 是否使用确定性算法
    """
    combined_seed = seed + rank
    random.seed(combined_seed)
    np.random.seed(combined_seed)
    torch.manual_seed(combined_seed)
    if torch.cuda.is_available():
        import torch.backends.cudnn as cudnn
        import torch.backends.cuda as cuda
        cuda.matmul.allow_tf32 = True
        cudnn.benchmark = True
        torch.cuda.manual_seed(combined_seed)
        torch.cuda.manual_seed_all(combined_seed)

    if torch.backends.flags_frozen():
        warnings.warn("PyTorch global flag support of backends is disabled, enable it to set global `cudnn` flags.")
        torch.backends.__allow_nonbracketed_mutation_flag = True
        
    if use_deterministic_algorithms:
        if hasattr(torch, "use_deterministic_algorithms"):
            torch.use_deterministic_algorithms(True)
        elif hasattr(torch, "set_deterministic"):
            torch.set_deterministic(True)
        else:
            warnings.warn("If use_deterministic-algorithms=True is set, pytorch version "
                          "greater than 1.8 may be required to use this feature properly.")

def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group['lr']



"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is used for outputting console information.
"""
import os
import re
import sys
import datetime
import argparse

def printlog(info):
    """Print log information with timestamp."""
    nowtime = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    print("\n"+"=========="*8 + "%s" % nowtime)
    print(info+'...\n\n')

def printprocess(info, index, nums):
    """打印进度信息，显示当前处理的进度"""
    print(f"\r[{info}] processing [{index}/{nums}], rate:{index/nums*100:.1f}%", end="")

def colorfulstr(obj, color="red", display_type="plain"):
    """
    Advanced version of "colorstr" function
    Args:
        obj: info content
        color: "black", "red", "green", "yellow", "blue", "purple", "cyan", "white"
        display_type: "plain", "highlight", "underline", "shine", "inverse", "invisible"

    Details:
        # 彩色输出格式：
        # 设置颜色开始 ：\033[显示方式;前景色;背景色m
        # 前景色            背景色            颜色
        # ---------------------------------------
        # 30                40              黑色
        # 31                41              红色
        # 32                42              绿色
        # 33                43              黃色
        # 34                44              蓝色
        # 35                45              紫红色
        # 36                46              青蓝色
        # 37                47              白色
        # 显示方式           意义
        # -------------------------
        # 0                终端默认设置
        # 1                高亮显示
        # 4                使用下划线
        # 5                闪烁
        # 7                反白显示
        # 8                不可见
    """
    color_dict = {"black": "30", "red": "31", "green": "32", "yellow": "33",
                  "blue": "34", "purple": "35","cyan": "36",  "white": "37"}
    display_type_dict = {"plain": "0", "highlight": "1", "underline": "4",
                         "shine": "5", "inverse": "7", "invisible": "8"}
    s = str(obj)
    color_code = color_dict.get(color, "")
    display = display_type_dict.get(display_type, "")
    out = '\033[{};{}m'.format(display, color_code) + s + '\033[0m'
    return out


def colorstr(*input):
    # https://en.wikipedia.org/wiki/ANSI_escape_code, i.e.  colorstr('blue', 'hello world')
    *args, string = input if len(input) > 1 else ('blue', 'bold', input[0])
    colors = {
        'black': '\033[30m',  # basic colors
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'magenta': '\033[35m',
        'cyan': '\033[36m',
        'white': '\033[37m',
        'bright_black': '\033[90m',  # bright colors
        'bright_red': '\033[91m',
        'bright_green': '\033[92m',
        'bright_yellow': '\033[93m',
        'bright_blue': '\033[94m',
        'bright_magenta': '\033[95m',
        'bright_cyan': '\033[96m',
        'bright_white': '\033[97m',
        'end': '\033[0m',  # misc
        'bold': '\033[1m',
        'underline': '\033[4m'}
    return ''.join(colors[x] for x in args) + f'{string}' + colors['end']


def show_config(head="Configurations", args=None, color='black', **kwargs):
    """显示配置信息"""
    print(colorstr(color, f'{head}:'))
    print(colorstr(color, '-' * 113))
    print(colorstr(color, '|%5s | %45s | %55s|' % ('order', 'keys', 'values')))
    print(colorstr(color, '-' * 113))
    counter = 0
    if args is not None:
        if isinstance(args, argparse.Namespace):
            config_dict = vars(args)
            for key, value in config_dict.items():
                counter += 1
                print(colorstr(color, f'|%5d | %45s | %55s|' % (counter, key, value)))
        elif isinstance(args, list):
            for arg in args:
                counter += 1
                print(colorstr(color, f'|%5d | %45s | {"": >55}|' % (counter, arg)))  # Assuming each element in the list is a key
        else:
            counter += 1
            print(colorstr(color, f'|%5d | %45s | %55s|' % (counter, "arg", args)))  # Assuming args is a single value

    for key, value in kwargs.items():
        counter += 1
        print(colorstr(color, f'|%5d | %45s | %55s|' % (counter, key, value)))

    print(colorstr(color, '-' * 113))

class LoadingBar:
    def __init__(self, length: int = 40):
        """Simple dynamic display bar
        Example:
        ```
            loading_bar = LoadingBar()
            for i in range(101):
                progress = i / 100.0
                bar_string = loading_bar(progress)
                print(f"\r[{bar_string}] {i}% ", end="")
                time.sleep(0.1)
            print("\nLoading Complete!")
        ```
        """
        self.length = length
        self.symbols = ['┈', '░', '▒', '▓']

    def __call__(self, progress: float) -> str:
        p = int(progress * self.length*4 + 0.5)
        d, r = p // 4, p % 4
        return '┠┈' + d * '█' + ((self.symbols[r]) + max(0, self.length-1-d) * '┈' if p < self.length*4 else '') + "┈┨"


class ConsoleLogger:
    def __init__(self, log_file, encoding='utf-8'):
        self.log_file = log_file
        self.clean(self.log_file)
        self.terminal = sys.stdout
        self.encoding = encoding

    def write(self, message):
        message_no_color = self.remove_ansi_colors(message)
        with open(self.log_file, 'a', encoding=self.encoding) as log:
            log.write(message_no_color)

        self.terminal.write(message)

    def clean(self,log_file):
        if os.path.exists(log_file):
            os.remove(log_file)

    def flush(self):
        # 为了兼容一些不支持flush的环境
        self.terminal.flush()

    @staticmethod
    def remove_ansi_colors(text):
        """
        去除 ANSI 颜色代码。
        """
        # 正则表达式匹配 ANSI 转义序列
        ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
        return ansi_escape.sub('', text)


def redirect_console(log_path='./out.log'):
    """
    将控制台输出重定向到指定文件。

    Args:
        log_path (str): 日志文件的路径。
    """
    logger = ConsoleLogger(log_path)
    sys.stdout = logger  # 将标准输出重定向到 Logger

if __name__ == "__main__":
    import platform
    import torch

    # 设置日志文件路径
    log_file = './output.log'

    redirect_console(log_file)
    print("This is a test message.")
    print("Another message to log.")

    total_tasks = 15
    task_info = "Task A"
    for i in range(1, total_tasks + 1):
        printprocess(task_info, i, total_tasks)
    print("\n任务完成！")

    text = "Hello, Colorful World!"
    printlog("Starting the main process.")

    colors = ["red", "green", "blue", "yellow", "purple", "cyan", "white"]
    display_types = ["plain", "highlight", "underline", "shine", "inverse", "invisible"]

    show_config(Python_vision=".".join(map(str, [sys.version_info.major, sys.version_info.minor, sys.version_info.micro])),
                CUDA=torch.version.cuda,
                CUDA_device=torch.cuda.get_device_name(0) if torch.cuda.is_available() else None,
                Torch_vision=torch.__version__,
                os_info=platform.system() + " " + platform.version(),
                color='red',)

    for color in colors:
        for display_type in display_types:
            colored_text = colorfulstr(text, color=color, display_type=display_type)
            print(f"Color: {color}, Display Type: {display_type} -> {colored_text}")


from .applytransformer import (
    Images,
    random_apply,
    random_order,
    random_choice,
    uniform_augment,
    Compose,
    RandomApply,
    RandomChoice,
    RandomOrder,
    UniformAugment
)
from .cvpiltransformer import (
    RandomHorizontalFlip,
    RandomVerticalFlip,
    AdjustBrightness,
    RandomAdjustBrightness,
    AdjustGamma,
    RandomAdjustGamma,
    ToHSV,
    RandomRotation,
    GaussianBlur,
    Grayscale,
    EqualizeHistogram,
    RandomEqualizeHistogram,
    Centerzoom,
    Resize,
    RandomCrop,
    CenterCrop,
    RandomResizeCrop,
    Pad,
    ResizePad,
    InvertColor,
    RandomInvertColor,
    AdjustSharpness,
    RandomAdjustSharpness,
    AdjustSaturation,
    RandomAdjustSaturation,
    ColorJitter,
    ToTensor,
    MeanStdNormalize
)

from .tvisiontansformer import (
    TvisionToTensor,
    TvisionNormalize,
    TvisionCenterCrop,
    TvisionRandomCrop,
    TvisionRandomResize,
    TvisionRandomContrast,
    TvisionRandomBrightness,
    TvisionRandomVerticalFlip,
    TvisionRandomHorizontalFlip,
)

from .normal import (
    imagenet_denormal,
    imagenet_normal,
    min_max_normal,
    z_score_normal,
    linear_normal,
    zero_centered_normal,
    Normalizer,
    denormalize
)


from .augments import (
    crop_image_by_2points,
    crop_image_by_1points,
    center_crop,
    five_crop,
    centerzoom,
    flip, vertical_flip, horizontal_flip,
    resize,
    resizepad,
    adjust_brightness_cv2, adjust_brightness_numpy,
    rotate,
    adjust_gamma,
    pad_margin,
    erase,
    enhance_hsv,
    hist_equalize,
    random_lighting,
    random_rotation, random_rot90,
    random_horizontal_flip,
    random_vertical_flip,
    random_crop,
    random_resize_crop,
    addnoisy,
    addfog,
    addfog_channels,
    Retinex
)


from .blur import (
    meanblur,
    medianblur,
    gaussianblur,
    bilateralblur,
    medianfilter,
    meanfilter,
    gaussian_kernel,
    gaussianfilter,
    bilateralfilter
)

from .transforms import *

from .contour import getContours, labelpoint, get_warp_perspective, SearchOutline,\
    drawOutline, gradientOutline

from .mask_ops import (
    convert_np,
    clip,
    unique,
    RectMask,
    BinaryImg,
    up_low,
    approximate,
    ceilfloor,
    bool2mask,
    cv_distance,
    chamfer,
    fast_marching,
    inpaint_defect,
    cvt8png,
    cvtMask,
    imageContentMove,
    count_zero,
    count_white,
    count_nonzero
)


from .adaptivepooling import adaptive_pool2d, adaptive_catavgmax_pool2d, adaptive_avgmax_pool2d,\
    AdaptiveCatAvgMaxPool2d, AdaptiveAvgMaxPool2d, FastAdaptiveAvgPool2d, SelectAdaptivePool2d
from .pool import MaxPool2d, AvgPool2d, StridedPool2d, MedianPool2d

from .haarwavelet import HWDownsampling
from .strippooling import StripPooling

# Downsampling processing module

__all__ = [
    "adaptive_pool2d",
    "adaptive_avgmax_pool2d",
    "adaptive_catavgmax_pool2d",
    "AdaptiveAvgMaxPool2d",
    "AdaptiveCatAvgMaxPool2d",
    "FastAdaptiveAvgPool2d",
    "SelectAdaptivePool2d",
    "MaxPool2d",
    "AvgPool2d",
    "HWDownsampling",
    "StridedPool2d",
    "StripPooling",
    "MedianPool2d"
]


from .colorspace import (
    grayscale,
    rgb2bgr,
    bgr2rgb,
    bgr2hsv,
    hsv2bgr,
    rgb2hsv,
    hsv2rgb
)

from .cvplot import OverlayPng, PutBoxText, AddText, ConvertBbox, DrawPolygon, CornerRect
from .core import Timer, FPS, Runcodes, timing, ColorFind, DetectVideoColor, DetectImageColor
from .printf import ConsoleLogger, redirect_console, colorstr, colorfulstr, \
    show_config, LoadingBar, printprocess, printlog

from .io import (
    imreader,
    imwriter,
    display,
    url2image,
    imattributes,
    imshowplt,
    VideoCap,
    StackedImagesV1,
    StackedImagesV2,
    Stackedplt,
    Stackedtorch,
    Mp4toGif,
)


"""
- author: Auorui (夏天是冰红茶)
- creation time:2022.10
- pyzjr is the Codebase accumulated by my python programming.
  At present, it is only for my personal use. If you want to
  use it, please contact me. Here are my email and WeChat.
- WeChat: zjricetea
- Email: zjricetea@gmail.com / 2165648225@qq.com
- Note: Currently still being updated, please refer to the latest version for any changes that may occur

    https://www.pepy.tech/projects/pyzjr
"""
from pyzjr.version import pyzjr_info, __version__
from pyzjr.utils import *
from pyzjr.visualize import *
from pyzjr.data import *
from pyzjr.nn import *
from pyzjr.measure import *
from pyzjr.Z import *
from pyzjr.augmentation import *







from .classification import (
    cls_matrix,
    BinaryConfusionMatrix,
    MulticlassConfusionMatrix,
    ConfusionMatrixs,
    ModelIndex,
    calculate_metrics,
)
from .segment import (
    Miou,
    Recall,
    Precision,
    F1Score,
    DiceCoefficient,
    Accuracy,
    SegmentationIndex,
    AIU
)
from .infer_testset import (
    calculate_area,
    mean_iou,
    auc_roc,
    accuracy,
    dice,
    kappa,
    InferTestset
)
from .utils import (
    iou, jaccard, IoU_utils,
    f_score, Fscore_utils,
    accuracy, Accuracy_utils,
    precision, Precision_utils,
    recall, Recall_utils
)


import re
import os
from pathlib import Path
from datetime import datetime

def timestr(use_simple_format=True):
    """Generate a formatted datetime string."""
    format = "%Y_%m_%d_%H_%M_%S" if use_simple_format else "%A_%d_%B_%Y_%Hh_%Mm_%Ss"
    return f"{datetime.now().strftime(format)}"

def alphabetstr(capital=False):
    """字符标签"""
    alphabet = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']
    uppercase_alphabet = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z']

    return uppercase_alphabet if capital else alphabet

def natural_key(st):
    """
    将字符串拆分成字母和数字块
    """
    return [int(text) if text.isdigit() else text.lower()
            for text in re.split('([0-9]+)', st)]

def natsorted(a):
    """
    手写自然排序
    >>> a = ['num9', 'num5', 'num2']
    >>> sorted_a = natsorted(a)
    ['num2', 'num5', 'num9']
    """
    return sorted(a, key=natural_key)

def list_dirs(dirs):
    """
    返回给定目录中的所有目录
    """
    currentdirs = [str(f) for f in Path(dirs).iterdir() if f.is_dir()]
    return natsorted(currentdirs), len(currentdirs)

def list_files(dirs):
    """
    返回给定目录中的所有文件
    """
    currentfiles = [
        str(f) for f in Path(dirs).iterdir()
        if f.is_file() and not f.name.startswith(".")
    ]
    return natsorted(currentfiles), len(currentfiles)

def list_names(file_dir):
    """获取指定文件夹下的文件列表，并按自然排序方式排序."""
    files = natsorted(os.listdir(file_dir))
    files_num = len(files)
    return files, files_num

if __name__=="__main__":
    paths = r'D:\PythonProject\pyzjrPyPi\pyzjr'
    dirs, nums = list_dirs(paths)
    print(dirs, nums)
    files, nums = list_files(paths)
    print(files, nums)
    files, nums = list_names(paths)
    print(files, nums)




"""
Copyright (c) 2022, Auorui.
All rights reserved.

Call torchvision for image enhancement
"""
import random
import torch
import torchvision
import numpy as np
from pyzjr.utils.randfun import rand
import torchvision.transforms.functional as tf
from pyzjr.augmentation.transforms.applytransformer import Images

class TvisionToTensor(object):
    def __call__(self, image, target):
        image = tf.to_tensor(image)
        target = torch.as_tensor(np.array(target), dtype=torch.int64)
        return image, target

def pad_if_smaller(img, size, fill=0):
    """如果图像最小边长小于给定size，则用数值fill进行padding"""
    min_size = min(img.size)
    if min_size < size:
        ow, oh = img.size
        padh = size - oh if oh < size else 0
        padw = size - ow if ow < size else 0
        img = tf.pad(img, [0, 0, padw, padh], fill=fill)
    return img

class TvisionRandomResize(object):
    def __init__(self, min_size, max_size=None):
        self.min_size = min_size
        if max_size is None:
            max_size = min_size
        self.max_size = max_size

    def __call__(self, image, target):
        size = random.randint(self.min_size, self.max_size)
        # 这里size传入的是int类型，所以是将图像的最小边长缩放到size大小
        image = tf.resize(image, [size])
        # 这里的interpolation注意下，在torchvision(0.9.0)以后才有InterpolationMode.NEAREST
        # 如果是之前的版本需要使用PIL.Image.NEAREST
        if torchvision.__version__>= "0.9.0":
            NEAREST = torchvision.transforms.InterpolationMode.NEAREST
        else:
            NEAREST = Images.NEAREST
        target = tf.resize(target, [size], interpolation=NEAREST)
        return image, target

class TvisionRandomHorizontalFlip(object):
    def __init__(self, flip_prob):
        self.flip_prob = flip_prob

    def __call__(self, image, target):
        if random.random() < self.flip_prob:
            image = tf.hflip(image)
            target = tf.hflip(target)
        return image, target

class TvisionRandomVerticalFlip(object):
    def __init__(self, flip_prob):
        self.flip_prob = flip_prob

    def __call__(self, image, target):
        if random.random() < self.flip_prob:
            image = tf.vflip(image)
            target = tf.vflip(target)
        return image, target

class TvisionRandomCrop(object):
    def __init__(self, size):
        self.size = size

    def __call__(self, image, target):
        image = pad_if_smaller(image, self.size)
        target = pad_if_smaller(target, self.size, fill=255)
        crop_params = torchvision.transforms.RandomCrop.get_params(image, (self.size, self.size))
        image = tf.crop(image, *crop_params)
        target = tf.crop(target, *crop_params)
        return image, target

class TvisionCenterCrop(object):
    def __init__(self, size):
        self.size = size

    def __call__(self, image, target):
        image = tf.center_crop(image, self.size)
        target = tf.center_crop(target, self.size)
        return image, target

class TvisionNormalize(object):
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, image, target):
        image = tf.normalize(image, mean=self.mean, std=self.std)
        return image, target

class TvisionRandomBrightness(object):
    def __init__(self, brightness_range=(0.5, 2.0)):
        self.brightness_range = brightness_range

    def __call__(self, image, target):
        brightness_factor = rand(self.brightness_range[0], self.brightness_range[1])
        image = tf.adjust_brightness(image, brightness_factor)

        return image, target

class TvisionRandomContrast(object):
    def __init__(self, contrast_range=(0.5, 2.0)):
        self.contrast_range = contrast_range

    def __call__(self, image, target):
        contrast_factor = rand(self.contrast_range[0], self.contrast_range[1])
        image = tf.adjust_contrast(image, contrast_factor)

        return image, target


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import PIL.Image as Image
    from pyzjr.augmentation.transforms.applytransformer import Compose
    image = Image.open(r'./VOCdevkit\VOC2007\JPEGImages\2007_000032.jpg')
    target = Image.open(r'./VOCdevkit\VOC2007\SegmentationClass\2007_000032.png')

    # 创建数据增强操作的组合
    transforms = Compose([
        TvisionRandomResize(min_size=256, max_size=512),
        TvisionRandomHorizontalFlip(flip_prob=0.5),
        TvisionRandomCrop(size=224),
        TvisionCenterCrop(size=200),
        TvisionToTensor(),

        TvisionRandomContrast(),
        TvisionRandomBrightness(),
        TvisionNormalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ])

    transformed_image, transformed_target = transforms(image, target)
    transformed_image = np.clip(transformed_image, 0, 1)
    plt.imshow(transformed_image.permute(1, 2, 0))
    plt.axis('off')
    plt.show()



import cv2
import numpy as np
from PIL import Image
from itertools import repeat

from pyzjr.utils.check import is_Iterable

def _ntuple(n):
    def parse(x):
        if is_Iterable(x):
            return x
        return tuple(repeat(x, n))
    return parse

to_1tuple = _ntuple(1)
to_2tuple = _ntuple(2)
to_3tuple = _ntuple(3)
to_4tuple = _ntuple(4)
to_ntuple = _ntuple

def pil2cv(pil_image):
    """将PIL图像转换为OpenCV图像"""
    if pil_image.mode == 'L':
        open_cv_image = np.array(pil_image)
    else:
        open_cv_image = cv2.cvtColor(np.array(pil_image), cv2.COLOR_RGB2BGR)
    return open_cv_image

def cv2pil(cv_image):
    """将OpenCV图像转换为PIL图像"""
    if cv_image.ndim == 2:
        pil_image = Image.fromarray(cv_image, mode='L')
    else:
        pil_image = Image.fromarray(cv2.cvtColor(cv_image, cv2.COLOR_BGR2RGB))
    return pil_image






import numpy as np
import cv2
import string
from skimage.filters import threshold_otsu
from skimage.morphology import skeletonize
from skimage import morphology, measure

from pyzjr.utils.check import is_gray_image
from pyzjr.augmentation.contour import SearchOutline
from pyzjr.visualize.cvplot import PutBoxText

def SkeletonMap(target):
    """
    获取骨架图的信息
    :param target: 目标图
    :return: 骨架图与一个数组，其中每一行表示一个非零元素的索引(y,x)，包括行索引和列索引
    """
    if target.ndim == 2 or target.shape[2] == 1:
        gray = target
    else:
        gray = cv2.cvtColor(target, cv2.COLOR_RGB2GRAY)
    thresh = threshold_otsu(target)
    binary = gray > thresh
    skimage = skeletonize(binary)
    skepoints = np.argwhere(skimage)
    skimage = skimage.astype(np.uint8)
    return skimage, skepoints


def incircle(img, contours_arr, color=(0, 0, 255)):
    """
    轮廓最大内切圆算法,所有轮廓当中的内切圆
    :param img: 单通道图像
    :param contours_arr: ndarry的轮廓, 建议使用cv2.findContours,
                        contours_arr, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
                        pyzjr也有SearchOutline可用,但要转换成ndarry的格式
    :example:
        contour = [np.array([point], dtype=np.int32) for point in contours]
        # 平铺的方法
        flatten_contours = np.concatenate([cnt.flatten() for cnt in contours_arr])
        flatten_contours = flatten_contours.reshape(-1, 2)
    :param color: 绘制内切圆颜色
    :return: 绘制在原图的内切圆,内切圆直径,绘制出的图像与轮廓直接差距一个像素，是因为cv2.circle的半径参数必须为int类型
    """
    if is_gray_image:
        result = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    raw_dist = np.zeros(img.shape, dtype=np.float32)
    letters = list(string.ascii_uppercase)
    label = {}
    k = 0
    for contours in contours_arr:
        for i in range(img.shape[0]):
            for j in range(img.shape[1]):
                raw_dist[i, j] = cv2.pointPolygonTest(contours, (j, i), True)
        min_val, max_val, _, max_dist_pt = cv2.minMaxLoc(raw_dist)
        if max_val > .5:
            label[letters[k]] = max_val * 2
            k += 1
        cv2.circle(result, max_dist_pt, int(max_val), color, 1, 1, 0)

    return result, label


def outcircle(img, contours_arr, color=(0, 255, 0)):
    """
    轮廓外切圆算法,画出所有轮廓的外切圆
    :param img: 单通道图像
    :param contours_arr: ndarry的轮廓, 建议使用cv2.findContours,
                        contours_arr, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
                        pyzjr也有SearchOutline可用,但要转换成ndarry的格式
    :example:
        contour = [np.array([point], dtype=np.int32) for point in contours]
    :param color: 绘制外切圆颜色
    :return:
    """
    radii = []
    result = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    for k, cnt in enumerate(contours_arr):
        (x, y), radius = cv2.minEnclosingCircle(cnt)
        radii.append([(x,y),radius])
        center = (int(x), int(y))
        radius = int(radius)
        cv2.circle(result, center, radius, color, 1)
        cv2.circle(result, center, 1, color, 1)

    return result, radii

def foreground_contour_length(binary_img, minArea=5):
    """
    计算前景的轮廓长度, 返回两个值, 每个轮廓的长度和总长度
    :param binary_img: 二值图
    :param minArea: 最小过滤面积
    """
    contours_xy = SearchOutline(binary_img)
    contour_lengths = []
    for contour in contours_xy:
        area = cv2.contourArea(contour)
        if area > minArea:
            length = cv2.arcLength(contour, False)
            contour_lengths.append(length)
    all_length = np.sum(contour_lengths)
    return contour_lengths, all_length


def get_each_crack_areas(mask, thresh, merge_threshold=3, area_threshold=50):
    """每条裂缝的面积,并用大写字母来进行标记"""
    connected_image = morphology.closing(thresh, morphology.disk(merge_threshold))
    labeled_image = measure.label(connected_image, connectivity=2)
    region_props = measure.regionprops(labeled_image)
    area = {}
    Bboxing = []
    crack_label = ord('A')
    for region in region_props:
        area_value = region.area
        if area_value >= area_threshold:
            minr, minc, maxr, maxc = region.bbox
            Bboxing.append([(minc, minr), (maxc, maxr)])
            PutBoxText(mask, [minc, minr, maxc, maxr], chr(crack_label), mode=0, fontsize=.5)
            if crack_label <= ord('Z'):
                area[chr(crack_label)] = area_value
                crack_label += 1
    return area, Bboxing, mask

if __name__=="__main__":
    from pyzjr.measure.skeleton_extraction import skeletoncv
    path = r'D:\PythonProject\pyzjrPyPi\models_img\1604.png'
    cv_image = cv2.imread(path)
    image = skeletoncv(path)
    # image = BinaryImg(cv_image)
    contour_lengths, all_length = foreground_contour_length(image)

    print(all_length)
    for i, length in enumerate(contour_lengths):
        print(f"Contour {i} length: {length}")


import numpy as np
import cv2
from pyzjr.measure.pixel import SkeletonMap
from pyzjr.utils.check import is_gray_image

# 裂缝类型
class CrackType():
    """直方图投影法推断裂缝类型, 能区分线性裂缝和网状裂缝"""
    def __init__(self, threshold=3, HWratio=10, Histratio=0.5, alpha=35, beta=55):
        """
        初始化分类裂缝的参数
        :param threshold: 阈值，用于分类裂缝的阈值
        :param HWratio: 高宽比，用于分类裂缝的高宽比阈值
        :param Histratio: 直方图比例，用于分类裂缝的直方图比例阈值
        :param alpha: 倾斜裂纹分类的较低阈值角度
        :param beta: 倾斜裂纹分类的上阈值角度
        """
        self.threshold = threshold
        self.HWratio = HWratio
        self.Histratio = Histratio
        self.types = {0: 'Horizontal',
                      1: 'Vertical',
                      2: 'Oblique',
                      3: 'Mesh'}
        self.alpha = alpha
        self.beta = beta

    def inference_minAreaRect(self, minAreaRect):
        """
        旋转矩形框长边与x轴的夹角.
        旋转角度 angle 是相对于图像水平方向的夹角，范围是 -90 到 +90 度.
        然而，一般情况下，我们习惯将角度定义为相对于 x 轴正方向的夹角，范围是 -180 到 +180 度.
        """
        w, h = minAreaRect[1]
        if w > h:
            angle = int(minAreaRect[2])
        else:
            angle = -(90 - int(minAreaRect[2]))
        return w, h, angle

    def classify(self, minAreaRect, skeleton_pts, HW):
        """
        针对当前crack instance，对其进行分类；
        主要利用了骨骼点双向投影直方图、旋转矩形框宽高比/角度；
        :param minAreaRect: 最小外接矩形框，[(cx, cy), (w, h), angle]；
        :param skeleton_pts: 骨骼点坐标；
        :param HW: 当前patch的高、宽；
        """
        H, W = HW
        w, h, angle = self.inference_minAreaRect(minAreaRect)
        if w / h < self.HWratio or h / w < self.HWratio:
            pts_y, pts_x = skeleton_pts[:, 0], skeleton_pts[:, 1]
            hist_x = np.histogram(pts_x, W)
            hist_y = np.histogram(pts_y, H)
            print("ss1",hist_x[0], hist_y[0])
            if self.hist_judge(hist_x[0]) and self.hist_judge(hist_y[0]):
                return 3

        return self.angle2cls(angle)

    def hist_judge(self, hist_v, esp=1e-5):
        less_than_T = np.count_nonzero((hist_v > 0) & (hist_v <= self.threshold))
        more_than_T = np.count_nonzero(hist_v > self.threshold)
        return more_than_T / (less_than_T + esp) > self.Histratio

    def angle2cls(self, angle):
        angle = abs(angle)
        assert 0 <= angle <= 90, "ERROR: The angle value exceeds the limit and should be between 0 and 90 degrees!"
        if angle < self.alpha:
            return 0
        elif self.alpha <= angle <= self.beta:
            return 2
        elif angle > self.beta:
            return 1
        else:
            return None

def _get_minAreaRect_information(mask):
    """
    从二值化掩膜图像中获取最小外接矩形的相关信息
    :param mask:二值化掩膜图像，包含目标区域的白色区域
    :return:最小外接矩形的信息，包括中心点坐标、宽高和旋转角度
    """
    if is_gray_image(mask):
        gray = mask
    else:
        gray = cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY)
    _, mask = cv2.threshold(gray, 127, 255, cv2.THRESH_BINARY)
    contours, hierarchy = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    contour_merge = np.vstack(contours)
    minAreaRect = cv2.minAreaRect(contour_merge)
    return minAreaRect

def infertype(mask):
    """推导裂缝类型"""
    crack = CrackType()
    H, W = mask.shape[:2]
    minAreaRect = _get_minAreaRect_information(mask)
    skeimage, skepoints = SkeletonMap(mask)
    result = crack.classify(minAreaRect, skepoints, HW=(H, W))
    crack_type = crack.types[result]
    return crack_type

if __name__=="__main__":
    from pyzjr.measure.skeleton_extraction import (
        skeletoncv, sketion_medial_axis, sketionio)
    path = r'D:\PythonProject\pyzjrPyPi\models_img\1604.png'
    image_cv = skeletoncv(path)
    print(image_cv.shape)
    print(type(image_cv))

    image_io = sketionio(path)
    print(image_io.shape)
    print(type(image_io))

    image_axis = sketion_medial_axis(path)
    print(image_axis.shape)
    print(type(image_axis))

    crack_type = infertype(image_axis)
    print(crack_type)
